import queue
from datetime import datetime, timezone

import sentry_sdk
from azure.cognitiveservices.speech.audio import (
    AudioStreamFormat,
    AudioStreamWaveFormat,
    PushAudioInputStream,
)
from loguru import logger

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import AzureTranscriberConfig, Transcription
from vocode.streaming.transcriber.base_transcriber import BaseThreadAsyncTranscriber
from vocode.utils.sentry_utils import CustomSentrySpans, sentry_create_span


class AzureTranscriber(BaseThreadAsyncTranscriber[AzureTranscriberConfig]):
    def __init__(
        self,
        transcriber_config: AzureTranscriberConfig,
    ):
        super().__init__(transcriber_config)

        format = None
        if self.transcriber_config.audio_encoding == AudioEncoding.LINEAR16:
            format = AudioStreamFormat(
                samples_per_second=self.transcriber_config.sampling_rate,
                wave_stream_format=AudioStreamWaveFormat.PCM,
            )

        elif self.transcriber_config.audio_encoding == AudioEncoding.MULAW:
            format = AudioStreamFormat(
                samples_per_second=self.transcriber_config.sampling_rate,
                wave_stream_format=AudioStreamWaveFormat.MULAW,
            )

        import azure.cognitiveservices.speech as speechsdk

        self.push_stream = PushAudioInputStream(format)

        config = speechsdk.audio.AudioConfig(stream=self.push_stream)

        speech_config = speechsdk.SpeechConfig(
            subscription=getenv("AZURE_SPEECH_KEY"),
            region=getenv("AZURE_SPEECH_REGION"),
        )

        speech_params = {
            "speech_config": speech_config,
            "audio_config": config,
        }

        if self.transcriber_config.candidate_languages:
            speech_config.set_property(
                property_id=speechsdk.PropertyId.SpeechServiceConnection_LanguageIdMode,
                value="Continuous",
            )
            auto_detect_source_language_config = (
                speechsdk.languageconfig.AutoDetectSourceLanguageConfig(
                    languages=self.transcriber_config.candidate_languages
                )
            )

            speech_params["auto_detect_source_language_config"] = auto_detect_source_language_config
        else:
            speech_params["language"] = self.transcriber_config.language

        self.speech = speechsdk.SpeechRecognizer(**speech_params)

        self._ended = False
        self.is_ready = False

    def recognized_sentence_final(self, evt):

        sentry_create_span(
            sentry_callable=sentry_sdk.start_span,
            op=CustomSentrySpans.LATENCY_OF_CONVERSATION,
            start_timestamp=datetime.now(tz=timezone.utc),
        )
        self.output_janus_queue.sync_q.put_nowait(
            Transcription(message=evt.result.text, confidence=1.0, is_final=True)
        )

    def recognized_sentence_stream(self, evt):
        self.output_janus_queue.sync_q.put_nowait(
            Transcription(message=evt.result.text, confidence=1.0, is_final=False)
        )

    def _run_loop(self):
        stream = self.generator()

        def stop_cb(evt):
            logger.debug("CLOSING on {}".format(evt))
            self.speech.stop_continuous_recognition()
            self._ended = True

        self.speech.recognizing.connect(lambda x: self.recognized_sentence_stream(x))
        self.speech.recognized.connect(lambda x: self.recognized_sentence_final(x))
        self.speech.session_started.connect(
            lambda evt: logger.debug("SESSION STARTED: {}".format(evt))
        )
        self.speech.session_stopped.connect(
            lambda evt: logger.debug("SESSION STOPPED {}".format(evt))
        )
        self.speech.canceled.connect(lambda evt: logger.debug("CANCELED {}".format(evt)))

        self.speech.session_stopped.connect(stop_cb)
        self.speech.canceled.connect(stop_cb)
        self.speech.start_continuous_recognition_async()

        for content in stream:
            self.push_stream.write(content)
            if self._ended:
                break

    def generator(self):
        while not self._ended:
            # Use a blocking get() to ensure there's at least one chunk of
            # data, and stop iteration if the chunk is None, indicating the
            # end of the audio stream.
            try:
                chunk = self.input_janus_queue.sync_q.get(timeout=5)
            except queue.Empty:
                return

            if chunk is None:
                return
            data = [chunk]

            # Now consume whatever other data's still buffered.
            while True:
                try:
                    chunk = self.input_janus_queue.sync_q.get_nowait()
                    if chunk is None:
                        return
                    data.append(chunk)
                except queue.Empty:
                    break

            yield b"".join(data)

    def terminate(self):
        self._ended = True
        self.speech.stop_continuous_recognition_async()
        super().terminate()



from typing import Optional

from loguru import logger

from vocode.streaming.utils.redis import initialize_redis_bytes
from vocode.streaming.utils.singleton import Singleton


class AudioCache(Singleton):
    def __init__(self):
        self.redis = initialize_redis_bytes()
        self.disabled = False

    @staticmethod
    async def safe_create():
        if AudioCache in Singleton._instances:
            return Singleton._instances[AudioCache]

        audio_cache = AudioCache()
        try:
            await audio_cache.redis.ping()
        except Exception:
            logger.warning("Redis ping failed on startup, disabling audio cache")
            audio_cache.disabled = True
        return audio_cache

    def get_audio_key(self, voice_identifier: str, text: str) -> str:
        return f"audio_cache:{voice_identifier}:{text}"

    async def get_audio(self, voice_identifier: str, text: str) -> Optional[bytes]:
        audio_key = self.get_audio_key(voice_identifier, text)
        if self.disabled:
            return None
        return await self.redis.get(audio_key)

    async def set_audio(
        self, voice_identifier: str, text: str, audio: bytes, ttl: Optional[int] = None
    ):
        # TODO: cache eviction
        if self.disabled:
            logger.warning("Audio cache is disabled")
            return
        logger.info(f"Setting audio for {voice_identifier} {text}")
        audio_key = self.get_audio_key(voice_identifier, text)
        await self.redis.set(audio_key, audio)
        if ttl is not None:
            await self.redis.expire(audio_key, ttl)



import asyncio
import sys
from concurrent.futures import ThreadPoolExecutor, wait
from typing import Optional

from loguru import logger

from vocode.turn_based.agent.base_agent import BaseAgent

raise DeprecationWarning("This Agent is deprecated and will be removed in the future.")


class StopThreadException(Exception):
    pass


class GPT4AllAgent(BaseAgent):
    SENTENCE_ENDINGS = [".", "!", "?"]

    DEFAULT_PROMPT_TEMPLATE = "{history}\nHuman: {human_input}\nAI:"

    def __init__(
        self,
        model_path: str,
        system_prompt: str,
        initial_message: Optional[str] = None,
    ):
        from pygpt4all.models.gpt4all_j import GPT4All_J

        super().__init__(initial_message)
        self.prompt_template = f"{system_prompt}\n\n{self.DEFAULT_PROMPT_TEMPLATE}"
        self.memory = [f"AI: {initial_message}"] if initial_message else []
        self.llm = GPT4All_J(model_path)
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    def create_prompt(self, human_input):
        history = "\n".join(self.memory[-5:])
        return self.prompt_template.format(history=history, human_input=human_input)

    def get_memory_entry(self, human_input, response):
        return f"Human: {human_input}\nAI: {response}"

    def respond(
        self,
        human_input,
    ) -> str:
        logger.debug("LLM responding to human input")
        prompt = self.create_prompt(human_input)
        response_buffer = ""

        def new_text_callback(text):
            nonlocal response_buffer
            response_buffer += text
            if len(response_buffer) > len(prompt) and response_buffer.endswith("Human:"):
                response_buffer = response_buffer[: -len("Human:")]
                sys.exit()

        future = self.thread_pool_executor.submit(
            self.llm.generate,
            prompt,
            new_text_callback=new_text_callback,
        )
        wait([future], timeout=10)
        response = response_buffer[(len(prompt) + 1) :]
        self.memory.append(self.get_memory_entry(human_input, response))
        logger.debug(f"LLM response: {response}")
        return response

    async def respond_async(self, human_input) -> str:
        prompt = self.create_prompt(human_input)
        response_buffer = ""

        def new_text_callback(text):
            nonlocal response_buffer
            response_buffer += text
            if len(response_buffer) > len(prompt) and response_buffer.endswith("Human:"):
                response_buffer = response_buffer[: -len("Human:")]
                raise StopThreadException("Stopping the thread")

        loop = asyncio.get_event_loop()
        try:
            await loop.run_in_executor(
                self.thread_pool_executor,
                lambda: self.llm.generate(prompt, new_text_callback=new_text_callback),
            )
        except StopThreadException:
            pass
        response = response_buffer[(len(prompt) + 1) :]
        self.memory.append(self.get_memory_entry(human_input, response))
        return response


if __name__ == "__main__":

    async def main():
        chat_responder = GPT4AllAgent(
            system_prompt="The AI is having a pleasant conversation about life.",
            model_path="/Users/ajayraj/Downloads/ggml-gpt4all-j-v1.3-groovy.bin",
        )
        while True:
            response = await chat_responder.respond_async(input("Human: "))
            print(f"AI: {response}")

    asyncio.run(main())



import asyncio
import io
import wave
from concurrent.futures import ThreadPoolExecutor
from typing import Any

import google.auth
from google.cloud import texttospeech as tts  # type: ignore

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import GoogleSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult


class GoogleSynthesizer(BaseSynthesizer[GoogleSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: GoogleSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        google.auth.default()

        # Instantiates a client
        self.client = tts.TextToSpeechClient()

        # Build the voice request, select the language code ("en-US") and the ssml
        # voice gender ("neutral")
        self.voice = tts.VoiceSelectionParams(
            language_code=synthesizer_config.language_code,
            name=synthesizer_config.voice_name,
        )

        # Select the type of audio file you want returned
        self.audio_config = tts.AudioConfig(
            audio_encoding=tts.AudioEncoding.LINEAR16,
            sample_rate_hertz=24000,
            speaking_rate=synthesizer_config.speaking_rate,
            pitch=synthesizer_config.pitch,
            effects_profile_id=["telephony-class-application"],
        )
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    def synthesize(self, message: str) -> Any:
        synthesis_input = tts.SynthesisInput(text=message)

        # Perform the text-to-speech request on the text input with the selected
        # voice parameters and audio file type
        return self.client.synthesize_speech(
            request=tts.SynthesizeSpeechRequest(
                input=synthesis_input,
                voice=self.voice,
                audio_config=self.audio_config,
                enable_time_pointing=[tts.SynthesizeSpeechRequest.TimepointType.SSML_MARK],
            )
        )

    # TODO: make this nonblocking, see speech.TextToSpeechAsyncClient
    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        response: tts.SynthesizeSpeechResponse = (  # type: ignore
            await asyncio.get_event_loop().run_in_executor(
                self.thread_pool_executor, self.synthesize, message.text
            )
        )
        output_sample_rate = response.audio_config.sample_rate_hertz

        output_bytes_io = io.BytesIO()
        in_memory_wav = wave.open(output_bytes_io, "wb")
        in_memory_wav.setnchannels(1)
        in_memory_wav.setsampwidth(2)
        in_memory_wav.setframerate(output_sample_rate)
        in_memory_wav.writeframes(response.audio_content)
        output_bytes_io.seek(0)

        result = self.create_synthesis_result_from_wav(
            synthesizer_config=self.synthesizer_config,
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )
        return result



from vocode.streaming.models.audio import AudioEncoding, SamplingRate

# TODO(EPD-186): namespace as Twilio
DEFAULT_SAMPLING_RATE = SamplingRate.RATE_8000
DEFAULT_AUDIO_ENCODING = AudioEncoding.MULAW
DEFAULT_CHUNK_SIZE = 20 * 160
MULAW_SILENCE_BYTE = b"\xff"

VONAGE_SAMPLING_RATE = SamplingRate.RATE_16000
VONAGE_AUDIO_ENCODING = AudioEncoding.LINEAR16
VONAGE_CHUNK_SIZE = 640  # 20ms at 16kHz with 16bit samples
VONAGE_CONTENT_TYPE = "audio/l16;rate=16000"
PCM_SILENCE_BYTE = b"\x00"



import os
from typing import Any, Dict, List, Optional

from vocode.streaming.models.telephony import VonageConfig
from vocode.streaming.telephony.client.abstract_telephony_client import AbstractTelephonyClient
from vocode.streaming.telephony.constants import VONAGE_CONTENT_TYPE
from vocode.streaming.utils.async_requester import AsyncRequestor


class VonageBadRequestException(ValueError):
    pass


class VonageClient(AbstractTelephonyClient):
    def __init__(
        self,
        base_url: str,
        maybe_vonage_config: Optional[VonageConfig] = None,
        record_calls: bool = False,
    ):

        import vonage

        self.vonage = vonage

        super().__init__(
            base_url=base_url,
        )
        self.vonage_config = maybe_vonage_config or VonageConfig(
            api_key=os.environ["VONAGE_API_KEY"],
            api_secret=os.environ["VONAGE_API_SECRET"],
            application_id=os.environ["VONAGE_APPLICATION_ID"],
            private_key=os.environ["VONAGE_PRIVATE_KEY"],
            record=record_calls,
        )
        # Vonage's sync client: only used for authentication helpers
        self.client = self.vonage.Client(
            key=self.vonage_config.api_key,
            secret=self.vonage_config.api_secret,
            application_id=self.vonage_config.application_id,
            private_key=self.vonage_config.private_key,
        )

    def get_telephony_config(self):
        return self.vonage_config

    async def create_call(
        self,
        conversation_id: str,
        to_phone: str,
        from_phone: str,
        record: bool = False,
        digits: Optional[str] = None,
        telephony_params: Optional[Dict[str, str]] = None,
    ) -> str:  # identifier of the call on the telephony provider
        return await self._create_vonage_call(
            to_phone,
            from_phone,
            self.create_call_ncco(
                conversation_id=conversation_id,
                record=record,
                is_outbound=True,
            ),
            digits,
            event_urls=[],
        )

    async def end_call(self, id) -> bool:
        async with AsyncRequestor().get_session().put(
            f"https://api.nexmo.com/v1/calls/{id}",
            json={"action": "hangup"},
            headers={"Authorization": f"Bearer {self.client._generate_application_jwt().decode()}"},
        ) as response:
            if not response.ok:
                raise RuntimeError(f"Failed to end call: {response.status} {response.reason}")
        return True

    async def update_call(self, vonage_uuid, new_ncco):
        async with AsyncRequestor().get_session().put(
            f"https://api.nexmo.com/v1/calls/{vonage_uuid}",
            json={
                "action": "transfer",
                "destination": {"type": "ncco", "ncco": new_ncco},
            },
            headers={"Authorization": self.client._create_jwt_auth_string().decode()},
        ) as response:
            if not response.ok:
                raise RuntimeError(f"Failed to update call: {response.status} {response.reason}")
            return True

    def create_call_ncco(
        self,
        conversation_id,
        record,  # currently no-op
        is_outbound: bool = False,
    ):
        ncco: List[Dict[str, Any]] = []
        ncco.append(
            {
                "action": "connect",
                "endpoint": [
                    {
                        "type": "websocket",
                        "uri": f"wss://{self.base_url}/connect_call/{conversation_id}",
                        "content-type": VONAGE_CONTENT_TYPE,
                        "headers": {},
                    }
                ],
            },
        )
        return ncco

    async def _create_vonage_call(
        self,
        to_phone: str,
        from_phone: str,
        ncco: str,
        digits: Optional[str] = None,
        event_urls: List[str] = [],
        **kwargs,
    ) -> str:  # returns the Vonage UUID
        vonage_call_uuid: str
        async with AsyncRequestor().get_session().post(
            "https://api.nexmo.com/v1/calls",
            json={
                "to": [{"type": "phone", "number": to_phone, "dtmfAnswer": digits}],
                "from": {"type": "phone", "number": from_phone},
                "ncco": ncco,
                "event_url": event_urls,
                **kwargs,
            },
            headers={"Authorization": f"Bearer {self.client._generate_application_jwt().decode()}"},
        ) as response:
            if not response.ok:
                if response.status == 400:
                    raise VonageBadRequestException(
                        "Failed to start call; this is usually due to a bad/malformed number. "
                        "If this persists, and you're sure that the number is well-formed, "
                        "please contact us."
                    )
                raise RuntimeError(f"Failed to start call: {response.status} {response.reason}")
            data = await response.json()
            if not data["status"] == "started":
                raise RuntimeError(f"Failed to start call: {response}")
            vonage_call_uuid = data["uuid"]
        return vonage_call_uuid

    async def send_dtmf(self, vonage_uuid: str, digits: str):
        async with AsyncRequestor().get_session().put(
            f"https://api.nexmo.com/v1/calls/{vonage_uuid}/dtmf",
            json={"digits": digits},
            headers={"Authorization": self.client._create_jwt_auth_string().decode()},
        ) as response:
            if not response.ok:
                raise RuntimeError(f"Failed to send DTMF: {response.status} {response.reason}")
            await response.json()



from typing import Type

from loguru import logger
from pydantic.v1 import BaseModel

from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.models.actions import ActionConfig as VocodeActionConfig
from vocode.streaming.models.actions import ActionInput, ActionOutput

_END_CONVERSATION_ACTION_DESCRIPTION = """
Ends the current conversation; use this when we should hang up. Use this
when: someone has indicated they expect the call to finish, and there are no further objectives
to complete. If the user says Goodbye or a different farewell phrase, this indicates that we should
hang up. Example farewell phrases: goodbye, so long, see you, later, take care, etc. People will
sometimes say a farewell phrase at the end of a longer sentence, so make sure to hang up then
as well. Example: \"This sounds great! Talk to you soon.\""""


class EndConversationParameters(BaseModel):
    pass


class EndConversationResponse(BaseModel):
    success: bool


class EndConversationVocodeActionConfig(
    VocodeActionConfig, type="action_end_conversation"  # type: ignore
):
    def action_attempt_to_string(self, input: ActionInput) -> str:
        assert isinstance(input.params, EndConversationParameters)
        return "Attempting to end conversation"

    def action_result_to_string(self, input: ActionInput, output: ActionOutput) -> str:
        assert isinstance(output.response, EndConversationResponse)
        if output.response.success:
            action_description = "Successfully ended conversation"
        else:
            action_description = "Did not end call because user interrupted"
        return action_description


class EndConversation(
    BaseAction[
        EndConversationVocodeActionConfig,
        EndConversationParameters,
        EndConversationResponse,
    ]
):
    description: str = _END_CONVERSATION_ACTION_DESCRIPTION
    parameters_type: Type[EndConversationParameters] = EndConversationParameters
    response_type: Type[EndConversationResponse] = EndConversationResponse

    def __init__(
        self,
        action_config: EndConversationVocodeActionConfig,
    ):
        super().__init__(
            action_config,
            quiet=True,
            should_respond="sometimes",
            is_interruptible=False,
        )

    async def _end_of_run_hook(self) -> None:
        """This method is called at the end of the run method. It is optional but intended to be
        overridden if needed."""
        pass

    async def run(
        self, action_input: ActionInput[EndConversationParameters]
    ) -> ActionOutput[EndConversationResponse]:
        if action_input.user_message_tracker is not None:
            await action_input.user_message_tracker.wait()

        if self.conversation_state_manager.transcript.was_last_message_interrupted():
            logger.info("Last bot message was interrupted")
            return ActionOutput(
                action_type=action_input.action_config.type,
                response=EndConversationResponse(success=False),
            )

        await self.conversation_state_manager.terminate_conversation()

        await self._end_of_run_hook()
        return ActionOutput(
            action_type=action_input.action_config.type,
            response=EndConversationResponse(success=True),
        )



import ctypes
import io
import pathlib
import wave
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from pydub import AudioSegment

from vocode.streaming.constants import SENTENCE_ENDINGS
from vocode.streaming.models.transcriber import Transcription, WhisperCPPTranscriberConfig
from vocode.streaming.transcriber.base_transcriber import BaseThreadAsyncTranscriber
from vocode.utils.whisper_cpp.helpers import transcribe
from vocode.utils.whisper_cpp.whisper_params import WhisperFullParams

WHISPER_CPP_SAMPLING_RATE = 16000


class WhisperCPPTranscriber(BaseThreadAsyncTranscriber[WhisperCPPTranscriberConfig]):
    def __init__(
        self,
        transcriber_config: WhisperCPPTranscriberConfig,
    ):
        super().__init__(transcriber_config)
        self._ended = False
        self.buffer_size = round(
            transcriber_config.sampling_rate * transcriber_config.buffer_size_seconds
        )
        self.buffer = np.empty(self.buffer_size, dtype=np.int16)
        self.buffer_index = 0

        # whisper cpp
        # load library and model
        libname = pathlib.Path().absolute() / self.transcriber_config.libname
        self.whisper = ctypes.CDLL(libname)  # type: ignore

        # tell Python what are the return types of the functions
        self.whisper.whisper_init_from_file.restype = ctypes.c_void_p
        self.whisper.whisper_full_default_params.restype = WhisperFullParams
        self.whisper.whisper_full_get_segment_text.restype = ctypes.c_char_p

        # initialize whisper.cpp context
        self.ctx = self.whisper.whisper_init_from_file(
            self.transcriber_config.fname_model.encode("utf-8")
        )

        # get default whisper parameters and adjust as needed
        self.params = self.whisper.whisper_full_default_params()
        self.params.print_realtime = False
        self.params.print_progress = False
        self.params.single_segment = True
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    def create_new_buffer(self):
        buffer = io.BytesIO()
        wav = wave.open(buffer, "wb")
        wav.setnchannels(1)
        wav.setsampwidth(2)
        wav.setframerate(self.transcriber_config.sampling_rate)
        return wav, buffer

    def _run_loop(self):
        in_memory_wav, audio_buffer = self.create_new_buffer()
        message_buffer = ""
        while not self._ended:
            chunk = self.input_janus_queue.sync_q.get()
            in_memory_wav.writeframes(chunk)
            if audio_buffer.tell() >= self.buffer_size * 2:
                audio_buffer.seek(0)
                audio_segment = AudioSegment.from_wav(audio_buffer)
                message, confidence = transcribe(self.whisper, self.params, self.ctx, audio_segment)
                message_buffer += message
                is_final = any(message_buffer.endswith(ending) for ending in SENTENCE_ENDINGS)
                in_memory_wav, audio_buffer = self.create_new_buffer()
                self.output_queue.put_nowait(
                    Transcription(message=message_buffer, confidence=confidence, is_final=is_final)
                )
                if is_final:
                    message_buffer = ""

    def terminate(self):
        pass



import re
from typing import List, Optional

_GOODBYE_PHRASES = [
    "bye",
]


def is_goodbye_simple(message: str, phrases: Optional[List[str]]):
    if not phrases:
        phrases = _GOODBYE_PHRASES
    cleaned = re.sub(r"[^\w\s]", "", message.lower())
    return any(phrase in cleaned for phrase in phrases)



import asyncio
import io
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from bark import SAMPLE_RATE, generate_audio, preload_models
from loguru import logger
from scipy.io.wavfile import write as write_wav

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import BarkSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult


class BarkSynthesizer(BaseSynthesizer[BarkSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: BarkSynthesizerConfig,
    ) -> None:
        super().__init__(synthesizer_config)

        self.SAMPLE_RATE = SAMPLE_RATE
        self.generate_audio = generate_audio
        logger.info("Loading Bark models")
        preload_models(**self.synthesizer_config.preload_kwargs)
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        logger.debug("Bark synthesizing audio")
        audio_array = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor,
            self.generate_audio,
            message.text,
            **self.synthesizer_config.generate_kwargs,
        )
        int_audio_arr = (audio_array * np.iinfo(np.int16).max).astype(np.int16)

        output_bytes_io = io.BytesIO()
        write_wav(output_bytes_io, self.SAMPLE_RATE, int_audio_arr)

        result = self.create_synthesis_result_from_wav(
            synthesizer_config=self.synthesizer_config,
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )

        return result



import os
from typing import Any, AsyncGenerator, Dict

import sentry_sdk
from anthropic import AsyncAnthropic, AsyncStream
from anthropic.types import MessageStreamEvent
from loguru import logger

from vocode.streaming.action.abstract_factory import AbstractActionFactory
from vocode.streaming.action.default_factory import DefaultActionFactory
from vocode.streaming.agent.anthropic_utils import format_anthropic_chat_messages_from_transcript
from vocode.streaming.agent.base_agent import GeneratedResponse, RespondAgent, StreamedResponse
from vocode.streaming.agent.streaming_utils import collate_response_async, stream_response_async
from vocode.streaming.models.actions import FunctionFragment
from vocode.streaming.models.agent import AnthropicAgentConfig
from vocode.streaming.models.message import BaseMessage, LLMToken
from vocode.streaming.vector_db.factory import VectorDBFactory
from vocode.utils.sentry_utils import CustomSentrySpans, sentry_create_span


class AnthropicAgent(RespondAgent[AnthropicAgentConfig]):
    anthropic_client: AsyncAnthropic

    def __init__(
        self,
        agent_config: AnthropicAgentConfig,
        action_factory: AbstractActionFactory = DefaultActionFactory(),
        vector_db_factory=VectorDBFactory(),
        **kwargs,
    ):
        super().__init__(
            agent_config=agent_config,
            action_factory=action_factory,
            **kwargs,
        )
        self.anthropic_client = AsyncAnthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))

    def get_chat_parameters(self, messages: list = [], use_functions: bool = True):
        assert self.transcript is not None

        parameters: dict[str, Any] = {
            "messages": messages,
            "system": self.agent_config.prompt_preamble,
            "max_tokens": self.agent_config.max_tokens,
            "temperature": self.agent_config.temperature,
            "stream": True,
        }

        parameters["model"] = self.agent_config.model_name

        return parameters

    async def token_generator(
        self,
        gen: AsyncStream[MessageStreamEvent],
    ) -> AsyncGenerator[str | FunctionFragment, None]:
        async for chunk in gen:
            if chunk.type == "content_block_delta" and chunk.delta.type == "text_delta":
                yield chunk.delta.text

    async def _get_anthropic_stream(self, chat_parameters: Dict[str, Any]):
        return await self.anthropic_client.messages.create(**chat_parameters)

    async def generate_response(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
        bot_was_in_medias_res: bool = False,
    ) -> AsyncGenerator[GeneratedResponse, None]:
        if not self.transcript:
            raise ValueError("A transcript is not attached to the agent")
        messages = format_anthropic_chat_messages_from_transcript(transcript=self.transcript)
        chat_parameters = self.get_chat_parameters(messages)
        try:
            first_sentence_total_span = sentry_create_span(
                sentry_callable=sentry_sdk.start_span, op=CustomSentrySpans.LLM_FIRST_SENTENCE_TOTAL
            )

            ttft_span = sentry_create_span(
                sentry_callable=sentry_sdk.start_span, op=CustomSentrySpans.TIME_TO_FIRST_TOKEN
            )
            stream = await self._get_anthropic_stream(chat_parameters)
        except Exception as e:
            logger.error(
                f"Error while hitting Anthropic with chat_parameters: {chat_parameters}",
                exc_info=True,
            )
            raise e

        response_generator = collate_response_async

        using_input_streaming_synthesizer = (
            self.conversation_state_manager.using_input_streaming_synthesizer()
        )
        if using_input_streaming_synthesizer:
            response_generator = stream_response_async
        async for message in response_generator(
            conversation_id=conversation_id,
            gen=self.token_generator(
                stream,
            ),
            sentry_span=ttft_span,
        ):
            if first_sentence_total_span:
                first_sentence_total_span.finish()

            ResponseClass = (
                StreamedResponse if using_input_streaming_synthesizer else GeneratedResponse
            )
            MessageType = LLMToken if using_input_streaming_synthesizer else BaseMessage

            if isinstance(message, str):
                yield ResponseClass(
                    message=MessageType(text=message),
                    is_interruptible=True,
                )
            else:
                yield ResponseClass(
                    message=message,
                    is_interruptible=True,
                )



from abc import abstractmethod
from typing import Literal, Optional, TypeVar, Union

from fastapi import WebSocket
from loguru import logger

from vocode import conversation_id as ctx_conversation_id
from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.events import PhoneCallEndedEvent
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import PhoneCallDirection
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.output_device.twilio_output_device import TwilioOutputDevice
from vocode.streaming.output_device.vonage_output_device import VonageOutputDevice
from vocode.streaming.streaming_conversation import StreamingConversation
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.utils import create_conversation_id
from vocode.streaming.utils.events_manager import EventsManager

TelephonyOutputDeviceType = TypeVar(
    "TelephonyOutputDeviceType", bound=Union[TwilioOutputDevice, VonageOutputDevice]
)

LOW_INTERRUPT_SENSITIVITY_THRESHOLD = 0.9

TelephonyProvider = Literal["twilio", "vonage"]


class AbstractPhoneConversation(StreamingConversation[TelephonyOutputDeviceType]):
    telephony_provider: TelephonyProvider

    def __init__(
        self,
        direction: PhoneCallDirection,
        from_phone: str,
        to_phone: str,
        base_url: str,
        config_manager: BaseConfigManager,
        output_device: TelephonyOutputDeviceType,
        agent_config: AgentConfig,
        transcriber_config: TranscriberConfig,
        synthesizer_config: SynthesizerConfig,
        agent_factory: AbstractAgentFactory,
        transcriber_factory: AbstractTranscriberFactory,
        synthesizer_factory: AbstractSynthesizerFactory,
        conversation_id: Optional[str] = None,
        events_manager: Optional[EventsManager] = None,
        speed_coefficient: float = 1.0,
        per_chunk_allowance_seconds: float = 0.01,
    ):
        conversation_id = conversation_id or create_conversation_id()
        ctx_conversation_id.set(conversation_id)

        self.direction = direction
        self.from_phone = from_phone
        self.to_phone = to_phone
        self.base_url = base_url
        super().__init__(
            output_device,
            transcriber_factory.create_transcriber(transcriber_config),
            agent_factory.create_agent(agent_config),
            synthesizer_factory.create_synthesizer(synthesizer_config),
            conversation_id=conversation_id,
            per_chunk_allowance_seconds=per_chunk_allowance_seconds,
            events_manager=events_manager,
            speed_coefficient=speed_coefficient,
        )
        self.transcriptions_worker = self.TranscriptionsWorker(
            input_queue=self.transcriber.output_queue,
            output_queue=self.agent.get_input_queue(),
            conversation=self,
            interruptible_event_factory=self.interruptible_event_factory,
        )
        self.config_manager = config_manager

    def attach_ws(self, ws: WebSocket):
        logger.debug("Trying to attach WS to outbound call")
        self.output_device.ws = ws
        logger.debug("Attached WS to outbound call")

    @abstractmethod
    async def attach_ws_and_start(self, ws: WebSocket):
        pass

    async def terminate(self):
        self.events_manager.publish_event(PhoneCallEndedEvent(conversation_id=self.id))
        await super().terminate()



import asyncio
from typing import Any, Dict, Optional

from vocode.streaming.action.base_action import ActionConfigType, BaseAction
from vocode.streaming.models.actions import (
    ActionInput,
    ParametersType,
    ResponseType,
    TwilioPhoneConversationActionInput,
    VonagePhoneConversationActionInput,
)
from vocode.streaming.utils.state_manager import (
    TwilioPhoneConversationStateManager,
    VonagePhoneConversationStateManager,
)


class VonagePhoneConversationAction(BaseAction[ActionConfigType, ParametersType, ResponseType]):
    def create_phone_conversation_action_input(
        self,
        conversation_id: str,
        params: Dict[str, Any],
        vonage_uuid: str,
        user_message_tracker: Optional[asyncio.Event] = None,
    ) -> VonagePhoneConversationActionInput[ParametersType]:
        if "user_message" in params:
            del params["user_message"]
        return VonagePhoneConversationActionInput(
            action_config=self.action_config,
            conversation_id=conversation_id,
            params=self.parameters_type(**params),
            vonage_uuid=vonage_uuid,
            user_message_tracker=user_message_tracker,
        )

    def get_vonage_uuid(self, action_input: ActionInput[ParametersType]) -> str:
        assert isinstance(action_input, VonagePhoneConversationActionInput)
        return action_input.vonage_uuid

    def attach_conversation_state_manager(self, conversation_state_manager: Any):
        assert isinstance(conversation_state_manager, VonagePhoneConversationStateManager)
        self.conversation_state_manager = conversation_state_manager


class TwilioPhoneConversationAction(BaseAction[ActionConfigType, ParametersType, ResponseType]):
    def create_phone_conversation_action_input(
        self,
        conversation_id: str,
        params: Dict[str, Any],
        twilio_sid: str,
        user_message_tracker: Optional[asyncio.Event] = None,
    ) -> TwilioPhoneConversationActionInput[ParametersType]:
        if "user_message" in params:
            del params["user_message"]
        return TwilioPhoneConversationActionInput(
            action_config=self.action_config,
            conversation_id=conversation_id,
            params=self.parameters_type(**params),
            twilio_sid=twilio_sid,
            user_message_tracker=user_message_tracker,
        )

    def get_twilio_sid(self, action_input: ActionInput[ParametersType]) -> str:
        assert isinstance(action_input, TwilioPhoneConversationActionInput)
        return action_input.twilio_sid

    def attach_conversation_state_manager(self, conversation_state_manager: Any):
        assert isinstance(conversation_state_manager, TwilioPhoneConversationStateManager)
        self.conversation_state_manager = conversation_state_manager



import os
from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union

import aiohttp
from openai import AsyncAzureOpenAI, AsyncOpenAI

from vocode.streaming.models.agent import AZURE_OPENAI_DEFAULT_API_VERSION

if TYPE_CHECKING:
    from langchain.docstore.document import Document

DEFAULT_OPENAI_EMBEDDING_MODEL = "text-embedding-ada-002"


class VectorDB:
    openai_client: Union[AsyncOpenAI, AsyncAzureOpenAI]

    def __init__(
        self,
        aiohttp_session: Optional[aiohttp.ClientSession] = None,
    ):
        if aiohttp_session:
            # the caller is responsible for closing the session
            self.aiohttp_session = aiohttp_session
            self.should_close_session_on_tear_down = False
        else:
            self.aiohttp_session = aiohttp.ClientSession()
            self.should_close_session_on_tear_down = True

        self.engine = os.getenv("AZURE_OPENAI_TEXT_EMBEDDING_ENGINE")
        if self.engine:
            azure_base = os.getenv("AZURE_OPENAI_API_BASE_EAST_US")
            azure_base = azure_base if azure_base is not None else ""
            self.openai_client = AsyncAzureOpenAI(
                azure_endpoint=azure_base,
                api_key=os.getenv("AZURE_OPENAI_API_KEY_EAST_US"),
                api_version=AZURE_OPENAI_DEFAULT_API_VERSION,
            )
        else:
            self.openai_client = AsyncOpenAI(
                api_key=os.getenv("OPENAI_API_KEY"),
            )

    async def create_openai_embedding(
        self, text, model=DEFAULT_OPENAI_EMBEDDING_MODEL
    ) -> List[float]:
        params = {
            "input": text,
        }

        params["model"] = self.engine if self.engine else model

        return (await self.openai_client.embeddings.create(**params)).data[0].embedding

    async def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        namespace: Optional[str] = None,
    ) -> List[str]:
        raise NotImplementedError

    async def similarity_search_with_score(
        self,
        query: str,
        filter: Optional[dict] = None,
        namespace: Optional[str] = None,
    ) -> List[Tuple["Document", float]]:
        raise NotImplementedError

    async def tear_down(self):
        if self.should_close_session_on_tear_down:
            await self.aiohttp_session.close()



from typing import Optional, Tuple, cast

import aiohttp
from loguru import logger

from vocode.streaming.agent.base_agent import RespondAgent
from vocode.streaming.models.agent import (
    RESTfulAgentInput,
    RESTfulAgentOutput,
    RESTfulAgentOutputType,
    RESTfulAgentText,
    RESTfulUserImplementedAgentConfig,
)


class RESTfulUserImplementedAgent(RespondAgent[RESTfulUserImplementedAgentConfig]):
    def __init__(
        self,
        agent_config: RESTfulUserImplementedAgentConfig,
    ):
        super().__init__(agent_config)
        if self.agent_config.generate_responses:
            raise NotImplementedError(
                "Use the WebSocket user implemented agent to stream responses"
            )

    async def respond(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
    ) -> Tuple[Optional[str], bool]:
        config = self.agent_config.respond
        try:
            # TODO: cache session
            async with aiohttp.ClientSession() as session:
                payload = RESTfulAgentInput(
                    human_input=human_input, conversation_id=conversation_id
                ).dict()
                async with session.request(
                    config.method,
                    config.url,
                    json=payload,
                    timeout=aiohttp.ClientTimeout(total=15),
                ) as response:
                    assert response.status == 200
                    output: RESTfulAgentOutput = RESTfulAgentOutput.parse_obj(await response.json())
                    output_response = None
                    should_stop = False
                    if output.type == RESTfulAgentOutputType.TEXT:
                        output_response = cast(RESTfulAgentText, output).response
                    elif output.type == RESTfulAgentOutputType.END:
                        should_stop = True
                    return output_response, should_stop
        except Exception as e:
            logger.error(f"Error in response from RESTful agent: {e}")
            return None, True



from __future__ import annotations

import asyncio
import base64
import json
from typing import Optional

from fastapi import WebSocket

from vocode.streaming.output_device.base_output_device import BaseOutputDevice
from vocode.streaming.telephony.constants import DEFAULT_AUDIO_ENCODING, DEFAULT_SAMPLING_RATE
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log


class TwilioOutputDevice(BaseOutputDevice):
    def __init__(self, ws: Optional[WebSocket] = None, stream_sid: Optional[str] = None):
        super().__init__(sampling_rate=DEFAULT_SAMPLING_RATE, audio_encoding=DEFAULT_AUDIO_ENCODING)
        self.ws = ws
        self.stream_sid = stream_sid
        self.active = True
        self.queue: asyncio.Queue[str] = asyncio.Queue()
        self.process_task = asyncio_create_task_with_done_error_log(self.process())

    async def process(self):
        while self.active:
            message = await self.queue.get()
            await self.ws.send_text(message)

    def consume_nonblocking(self, chunk: bytes):
        twilio_message = {
            "event": "media",
            "streamSid": self.stream_sid,
            "media": {"payload": base64.b64encode(chunk).decode("utf-8")},
        }
        self.queue.put_nowait(json.dumps(twilio_message))

    def send_chunk_finished_mark(self, utterance_id, chunk_idx):
        mark_message = {
            "event": "mark",
            "streamSid": self.stream_sid,
            "mark": {
                "name": f"chunk-{utterance_id}-{chunk_idx}",
            },
        }
        self.queue.put_nowait(json.dumps(mark_message))

    def send_utterance_finished_mark(self, utterance_id):
        mark_message = {
            "event": "mark",
            "streamSid": self.stream_sid,
            "mark": {
                "name": f"utterance-{utterance_id}",
            },
        }
        self.queue.put_nowait(json.dumps(mark_message))

    def send_clear_message(self):
        clear_message = {
            "event": "clear",
            "streamSid": self.stream_sid,
        }
        self.queue.put_nowait(json.dumps(clear_message))

    def terminate(self):
        self.process_task.cancel()



from enum import Enum
from typing import Optional

from vocode.streaming.models.agent import AgentConfig, AgentType
from vocode.streaming.models.model import BaseModel, TypedModel


class WebSocketAgentMessageType(str, Enum):
    BASE = "websocket_agent_base"
    TEXT = "websocket_agent_text"
    STOP = "websocket_agent_stop"


class WebSocketAgentMessage(TypedModel, type=WebSocketAgentMessageType.BASE):  # type: ignore
    conversation_id: Optional[str] = None


class WebSocketAgentTextMessage(
    WebSocketAgentMessage, type=WebSocketAgentMessageType.TEXT  # type: ignore
):
    class Payload(BaseModel):
        text: str

    data: Payload

    @classmethod
    def from_text(cls, text: str, conversation_id: Optional[str] = None):
        return cls(data=cls.Payload(text=text), conversation_id=conversation_id)


class WebSocketAgentStopMessage(
    WebSocketAgentMessage, type=WebSocketAgentMessageType.STOP  # type: ignore
):
    pass


class WebSocketUserImplementedAgentConfig(
    AgentConfig, type=AgentType.WEBSOCKET_USER_IMPLEMENTED.value  # type: ignore
):
    class RouteConfig(BaseModel):
        url: str

    respond: RouteConfig



from typing import Optional

from fastapi import APIRouter, HTTPException, WebSocket
from loguru import logger

from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.agent.default_factory import DefaultAgentFactory
from vocode.streaming.models.telephony import BaseCallConfig, TwilioCallConfig, VonageCallConfig
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.synthesizer.default_factory import DefaultSynthesizerFactory
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.telephony.conversation.abstract_phone_conversation import (
    AbstractPhoneConversation,
)
from vocode.streaming.telephony.conversation.twilio_phone_conversation import (
    TwilioPhoneConversation,
)
from vocode.streaming.telephony.conversation.vonage_phone_conversation import (
    VonagePhoneConversation,
)
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.transcriber.default_factory import DefaultTranscriberFactory
from vocode.streaming.utils.base_router import BaseRouter
from vocode.streaming.utils.events_manager import EventsManager


class CallsRouter(BaseRouter):
    def __init__(
        self,
        base_url: str,
        config_manager: BaseConfigManager,
        transcriber_factory: AbstractTranscriberFactory = DefaultTranscriberFactory(),
        agent_factory: AbstractAgentFactory = DefaultAgentFactory(),
        synthesizer_factory: AbstractSynthesizerFactory = DefaultSynthesizerFactory(),
        events_manager: Optional[EventsManager] = None,
    ):
        super().__init__()
        self.base_url = base_url
        self.config_manager = config_manager
        self.transcriber_factory = transcriber_factory
        self.agent_factory = agent_factory
        self.synthesizer_factory = synthesizer_factory
        self.events_manager = events_manager
        self.router = APIRouter()
        self.router.websocket("/connect_call/{id}")(self.connect_call)

    def _from_call_config(
        self,
        base_url: str,
        call_config: BaseCallConfig,
        config_manager: BaseConfigManager,
        conversation_id: str,
        transcriber_factory: AbstractTranscriberFactory = DefaultTranscriberFactory(),
        agent_factory: AbstractAgentFactory = DefaultAgentFactory(),
        synthesizer_factory: AbstractSynthesizerFactory = DefaultSynthesizerFactory(),
        events_manager: Optional[EventsManager] = None,
    ) -> AbstractPhoneConversation:
        if isinstance(call_config, TwilioCallConfig):
            return TwilioPhoneConversation(
                to_phone=call_config.to_phone,
                from_phone=call_config.from_phone,
                base_url=base_url,
                config_manager=config_manager,
                agent_config=call_config.agent_config,
                transcriber_config=call_config.transcriber_config,
                synthesizer_config=call_config.synthesizer_config,
                twilio_config=call_config.twilio_config,
                twilio_sid=call_config.twilio_sid,
                conversation_id=conversation_id,
                transcriber_factory=transcriber_factory,
                agent_factory=agent_factory,
                synthesizer_factory=synthesizer_factory,
                events_manager=events_manager,
                direction=call_config.direction,
            )
        elif isinstance(call_config, VonageCallConfig):
            return VonagePhoneConversation(
                to_phone=call_config.to_phone,
                from_phone=call_config.from_phone,
                base_url=base_url,
                config_manager=config_manager,
                agent_config=call_config.agent_config,
                transcriber_config=call_config.transcriber_config,
                synthesizer_config=call_config.synthesizer_config,
                vonage_config=call_config.vonage_config,
                vonage_uuid=call_config.vonage_uuid,
                conversation_id=conversation_id,
                transcriber_factory=transcriber_factory,
                agent_factory=agent_factory,
                synthesizer_factory=synthesizer_factory,
                events_manager=events_manager,
                output_to_speaker=call_config.output_to_speaker,
                direction=call_config.direction,
            )
        else:
            raise ValueError(f"Unknown call config type {call_config.type}")

    async def connect_call(self, websocket: WebSocket, id: str):
        await websocket.accept()
        logger.debug("Phone WS connection opened for chat {}".format(id))
        call_config = await self.config_manager.get_config(id)
        if not call_config:
            raise HTTPException(status_code=400, detail="No active phone call")

        phone_conversation = self._from_call_config(
            base_url=self.base_url,
            call_config=call_config,
            config_manager=self.config_manager,
            conversation_id=id,
            transcriber_factory=self.transcriber_factory,
            agent_factory=self.agent_factory,
            synthesizer_factory=self.synthesizer_factory,
            events_manager=self.events_manager,
        )

        await phone_conversation.attach_ws_and_start(websocket)
        logger.debug("Phone WS connection closed for chat {}".format(id))

    def get_router(self) -> APIRouter:
        return self.router



import uuid
from typing import Iterable, List, Optional, Tuple

from langchain.docstore.document import Document
from loguru import logger

from vocode import getenv
from vocode.streaming.models.vector_db import PineconeConfig
from vocode.streaming.vector_db.base_vector_db import VectorDB


class PineconeDB(VectorDB):
    def __init__(self, config: PineconeConfig, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.config = config

        self.index_name = self.config.index
        self.pinecone_api_key = getenv("PINECONE_API_KEY") or self.config.api_key
        self.pinecone_environment = getenv("PINECONE_ENVIRONMENT") or self.config.api_environment
        self.pinecone_url = f"https://{self.index_name}.svc.{self.pinecone_environment}.pinecone.io"
        self._text_key = "text"

    async def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        namespace: Optional[str] = None,
    ) -> List[str]:
        """Run more texts through the embeddings and add to the vectorstore.

        Args:
            texts: Iterable of strings to add to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            ids: Optional list of ids to associate with the texts.
            namespace: Optional pinecone namespace to add the texts to.

        Returns:
            List of ids from adding the texts into the vectorstore.
        """
        # Adapted from: langchain/vectorstores/pinecone.py. Made langchain implementation async.
        if namespace is None:
            namespace = ""
        # Embed and create the documents
        docs = []
        ids = ids or [str(uuid.uuid4()) for _ in texts]
        for i, text in enumerate(texts):
            embedding = await self.create_openai_embedding(text)
            metadata = metadatas[i] if metadatas else {}
            metadata[self._text_key] = text
            docs.append({"id": ids[i], "values": embedding, "metadata": metadata})
        # upsert to Pinecone
        async with self.aiohttp_session.post(
            f"{self.pinecone_url}/vectors/upsert",
            headers={"Api-Key": self.pinecone_api_key},
            json={
                "vectors": docs,
                "namespace": namespace,
            },
        ) as response:
            response_json = await response.json()
            if "message" in response_json:
                logger.error(f"Error upserting vectors: {response_json}")

        return ids

    async def similarity_search_with_score(
        self,
        query: str,
        filter: Optional[dict] = None,
        namespace: Optional[str] = None,
    ) -> List[Tuple[Document, float]]:
        """Return pinecone documents most similar to query, along with scores.

        Args:
            query: Text to look up documents similar to.
            filter: Dictionary of argument(s) to filter on metadata
            namespace: Namespace to search in. Default will search in '' namespace.

        Returns:
            List of Documents most similar to the query and score for each
        """
        # Adapted from: langchain/vectorstores/pinecone.py. Made langchain implementation async.
        if namespace is None:
            namespace = ""
        query_obj = await self.create_openai_embedding(query)
        docs = []
        async with self.aiohttp_session.post(
            f"{self.pinecone_url}/query",
            headers={"Api-Key": self.pinecone_api_key},
            json={
                "top_k": self.config.top_k,
                "namespace": namespace,
                "filter": filter,
                "vector": query_obj,
                "includeMetadata": True,
            },
        ) as response:
            results = await response.json()

        for res in results["matches"]:
            metadata = res["metadata"]
            if self._text_key in metadata:
                text = metadata.pop(self._text_key)
                score = res["score"]
                docs.append((Document(page_content=text, metadata=metadata), score))
            else:
                logger.warning(f"Found document with no `{self._text_key}` key. Skipping.")
        return docs



import asyncio
import io
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from pydub import AudioSegment
from TTS.api import TTS

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import CoquiTTSSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult


class CoquiTTSSynthesizer(BaseSynthesizer[CoquiTTSSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: CoquiTTSSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        self.tts = TTS(**synthesizer_config.tts_kwargs)
        self.speaker = synthesizer_config.speaker
        self.language = synthesizer_config.language
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        tts = self.tts
        audio_data = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor,
            tts.tts,
            message.text,
            self.speaker,
            self.language,
        )
        audio_data = np.array(audio_data)

        # Convert the NumPy array to bytes
        audio_data_bytes = (audio_data * 32767).astype(np.int16).tobytes()

        # Create an in-memory file-like object (BytesIO) to store the audio data
        buffer = io.BytesIO(audio_data_bytes)

        audio_segment: AudioSegment = AudioSegment.from_raw(
            buffer, frame_rate=22050, channels=1, sample_width=2  # type: ignore
        )

        output_bytes_io = io.BytesIO()
        audio_segment.export(output_bytes_io, format="wav")  # type: ignore

        result = self.create_synthesis_result_from_wav(
            synthesizer_config=self.synthesizer_config,
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )

        return result



from vocode.streaming.models.transcriber import (
    AssemblyAITranscriberConfig,
    AzureTranscriberConfig,
    DeepgramTranscriberConfig,
    GladiaTranscriberConfig,
    GoogleTranscriberConfig,
    RevAITranscriberConfig,
    TranscriberConfig,
)
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.transcriber.assembly_ai_transcriber import AssemblyAITranscriber
from vocode.streaming.transcriber.azure_transcriber import AzureTranscriber
from vocode.streaming.transcriber.deepgram_transcriber import DeepgramTranscriber
from vocode.streaming.transcriber.gladia_transcriber import GladiaTranscriber
from vocode.streaming.transcriber.google_transcriber import GoogleTranscriber
from vocode.streaming.transcriber.rev_ai_transcriber import RevAITranscriber


class DefaultTranscriberFactory(AbstractTranscriberFactory):
    def create_transcriber(
        self,
        transcriber_config: TranscriberConfig,
    ):
        if isinstance(transcriber_config, DeepgramTranscriberConfig):
            return DeepgramTranscriber(transcriber_config)
        elif isinstance(transcriber_config, GoogleTranscriberConfig):
            return GoogleTranscriber(transcriber_config)
        elif isinstance(transcriber_config, AssemblyAITranscriberConfig):
            return AssemblyAITranscriber(transcriber_config)
        elif isinstance(transcriber_config, RevAITranscriberConfig):
            return RevAITranscriber(transcriber_config)
        elif isinstance(transcriber_config, AzureTranscriberConfig):
            return AzureTranscriber(transcriber_config)
        elif isinstance(transcriber_config, GladiaTranscriberConfig):
            return GladiaTranscriber(transcriber_config)
        else:
            raise Exception("Invalid transcriber config")



import phonenumbers
from phonenumbers import PhoneNumberFormat

NUMBER_PARSE_ERROR = "Input number cannot be interpreted as a phone number. If passing an international number, please include the country code (e.g. +49 for German numbers)"


def parse_number_e164(phone_number: str) -> phonenumbers.PhoneNumber:
    if not phone_number.startswith("+"):
        phone_number = "+" + phone_number
    return phonenumbers.parse(phone_number, None)


def parse_number_usa_fallback_to_e164(phone_number: str) -> phonenumbers.PhoneNumber:
    try:
        phone_number_obj = phonenumbers.parse(phone_number, "US")
        valid = phonenumbers.is_valid_number(phone_number_obj)
        return phone_number_obj if valid else parse_number_e164(phone_number)
    except phonenumbers.phonenumberutil.NumberParseException:
        return parse_number_e164(phone_number)


def parse_phone_number(phone_number: str) -> phonenumbers.PhoneNumber:
    if phone_number.startswith("+"):
        # If we have a plus we know its e164
        return parse_number_e164(phone_number)
    else:
        # If we don't have a plus, we try as USA with e164 fallback
        return parse_number_usa_fallback_to_e164(phone_number)


def sanitize_phone_number(phone_number: str) -> str:
    phone_number_obj: phonenumbers.PhoneNumber

    try:
        phone_number_obj = parse_phone_number(phone_number)
    except Exception:
        raise ValueError(NUMBER_PARSE_ERROR)

    if not phonenumbers.is_valid_number(phone_number_obj):
        raise ValueError(NUMBER_PARSE_ERROR)

    return phonenumbers.format_number(phone_number_obj, PhoneNumberFormat.E164).replace("+", "")



from abc import ABC, abstractmethod
from typing import Dict, Optional

from vocode.streaming.models.telephony import TelephonyProviderConfig


class AbstractTelephonyClient(ABC):
    def __init__(self, base_url: str):
        self.base_url = base_url

    @abstractmethod
    def get_telephony_config(self) -> TelephonyProviderConfig:
        pass

    @abstractmethod
    async def create_call(
        self,
        conversation_id: str,
        to_phone: str,
        from_phone: str,
        record: bool = False,
        digits: Optional[str] = None,
        telephony_params: Optional[Dict[str, str]] = None,
    ) -> str:  # returns identifier of the call on the telephony provider
        pass

    @abstractmethod
    async def end_call(self, id) -> bool:
        raise NotImplementedError



import asyncio
import hashlib
from typing import Optional

from elevenlabs import Voice, VoiceSettings
from elevenlabs.client import AsyncElevenLabs
from loguru import logger

from vocode.streaming.models.audio import AudioEncoding, SamplingRate
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import ElevenLabsSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log

ELEVEN_LABS_BASE_URL = "https://api.elevenlabs.io/v1/"
STREAMED_CHUNK_SIZE = 16000 * 2 // 4  # 1/8 of a second of 16kHz audio with 16-bit samples


class ElevenLabsSynthesizer(BaseSynthesizer[ElevenLabsSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: ElevenLabsSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        assert synthesizer_config.api_key is not None, "API key must be set"
        assert synthesizer_config.voice_id is not None, "Voice ID must be set"
        self.api_key = synthesizer_config.api_key

        self.elevenlabs_client = AsyncElevenLabs(
            api_key=self.api_key,
        )

        self.voice_id = synthesizer_config.voice_id
        self.stability = synthesizer_config.stability
        self.similarity_boost = synthesizer_config.similarity_boost
        self.model_id = synthesizer_config.model_id
        self.optimize_streaming_latency = synthesizer_config.optimize_streaming_latency
        self.words_per_minute = 150
        self.upsample = None
        self.sample_rate = self.synthesizer_config.sampling_rate

        if self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            match self.synthesizer_config.sampling_rate:
                case SamplingRate.RATE_16000:
                    self.output_format = "pcm_16000"
                case SamplingRate.RATE_22050:
                    self.output_format = "pcm_22050"
                case SamplingRate.RATE_24000:
                    self.output_format = "pcm_24000"
                case SamplingRate.RATE_44100:
                    self.output_format = "pcm_44100"
                case SamplingRate.RATE_48000:
                    self.output_format = "pcm_44100"
                    self.upsample = SamplingRate.RATE_48000.value
                    self.sample_rate = SamplingRate.RATE_44100.value
                case _:
                    raise ValueError(
                        f"Unsupported sampling rate: {self.synthesizer_config.sampling_rate}. Elevenlabs only supports 16000, 22050, 24000, and 44100 Hz."
                    )
        elif self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            self.output_format = "ulaw_8000"
        else:
            raise ValueError(
                f"Unsupported audio encoding: {self.synthesizer_config.audio_encoding}"
            )

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        self.total_chars += len(message.text)
        if self.stability is not None and self.similarity_boost is not None:
            voice = Voice(
                voice_id=self.voice_id,
                settings=VoiceSettings(
                    stability=self.stability, similarity_boost=self.similarity_boost
                ),
            )
        else:
            voice = Voice(voice_id=self.voice_id)
        url = (
            ELEVEN_LABS_BASE_URL
            + f"text-to-speech/{self.voice_id}/stream?output_format={self.output_format}"
        )
        if self.optimize_streaming_latency:
            url += f"&optimize_streaming_latency={self.optimize_streaming_latency}"
        headers = {"xi-api-key": self.api_key}
        body = {
            "text": message.text,
            "voice_settings": voice.settings.dict() if voice.settings else None,
        }
        if self.model_id:
            body["model_id"] = self.model_id

        chunk_queue: asyncio.Queue[Optional[bytes]] = asyncio.Queue()
        asyncio_create_task_with_done_error_log(
            self.get_chunks(url, headers, body, chunk_size, chunk_queue),
        )

        return SynthesisResult(
            self.chunk_result_generator_from_queue(chunk_queue),
            lambda seconds: self.get_message_cutoff_from_voice_speed(message, seconds, 150),
        )

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: ElevenLabsSynthesizerConfig):
        hashed_api_key = hashlib.sha256(f"{synthesizer_config.api_key}".encode("utf-8")).hexdigest()
        return ":".join(
            (
                "eleven_labs",
                hashed_api_key,
                str(synthesizer_config.voice_id),
                str(synthesizer_config.model_id),
                str(synthesizer_config.stability),
                str(synthesizer_config.similarity_boost),
                str(synthesizer_config.optimize_streaming_latency),
                synthesizer_config.audio_encoding,
            )
        )

    async def get_chunks(
        self,
        url: str,
        headers: dict,
        body: dict,
        chunk_size: int,
        chunk_queue: asyncio.Queue[Optional[bytes]],
    ):
        try:
            async_client = self.async_requestor.get_client()
            stream = await async_client.send(
                async_client.build_request(
                    "POST",
                    url,
                    headers=headers,
                    json=body,
                ),
                stream=True,
            )

            if not stream.is_success:
                error = await stream.aread()
                logger.error(f"ElevenLabs API failed: {stream.status_code} {error.decode('utf-8')}")
                raise Exception(f"ElevenLabs API returned {stream.status_code} status code")
            async for chunk in stream.aiter_bytes(chunk_size):
                if self.upsample:
                    chunk = self._resample_chunk(
                        chunk,
                        self.sample_rate,
                        self.upsample,
                    )
                chunk_queue.put_nowait(chunk)
        except asyncio.CancelledError:
            pass
        finally:
            chunk_queue.put_nowait(None)  # treated as sentinel



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Tuple

from langchain import ConversationChain
from langchain.memory import ConversationBufferMemory
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate, MessagesPlaceholder
from langchain.schema import SystemMessage
from langchain_community.chat_models import ChatVertexAI
from loguru import logger

from vocode.streaming.agent.base_agent import RespondAgent
from vocode.streaming.models.agent import ChatVertexAIAgentConfig


class ChatVertexAIAgent(RespondAgent[ChatVertexAIAgentConfig]):
    def __init__(
        self,
        agent_config: ChatVertexAIAgentConfig,
    ):
        super().__init__(agent_config=agent_config)

        self.prompt = ChatPromptTemplate.from_messages(
            [
                MessagesPlaceholder(variable_name="history"),
                HumanMessagePromptTemplate.from_template("{input}"),
            ]
        )

        self.llm = ChatVertexAI()

        self.memory = ConversationBufferMemory(return_messages=True)
        self.memory.chat_memory.messages.append(
            SystemMessage(content=self.agent_config.prompt_preamble)
        )

        self.conversation = ConversationChain(memory=self.memory, prompt=self.prompt, llm=self.llm)
        if agent_config.initial_message:
            raise NotImplementedError("initial_message not supported for Vertex AI")
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    async def respond(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
    ) -> Tuple[str, bool]:
        # Vertex AI doesn't allow async, so we run in a separate thread
        text = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor,
            lambda input: self.conversation.predict(input=input),
            human_input,
        )

        logger.debug(f"LLM response: {text}")
        return text, False



import base64
import hashlib
import hmac
import json
from typing import Any, Dict, Optional

import httpx
from loguru import logger
from pydantic.v1 import BaseModel


class ExternalActionValueError(ValueError):
    pass


class ExternalActionsErrorResponses:
    client_error = (
        "There was an error with the information provided. "
        "Please ask for clarification and try again."
    )
    server_error = (
        "There was a server error with status "
        "{status}: {text}. Please politely tell the user to try again later."
    )
    input_error = (
        "There was an error with the information received from the service. "
        "Please politely tell the user to try again later."
        "\nThe error was {error}"
    )
    unauthorized = (
        "There was an error with the authentication for the service. "
        "Please politely tell the user to try again later."
    )
    forbidden = (
        "There was an error with the authentication for the service. "
        "Please politely tell the user to try again later."
    )


class ExternalActionResponse(BaseModel):
    result: dict
    success: bool
    agent_message: Optional[str] = None


class ExternalActionsRequester:
    def __init__(self, url: str) -> None:
        self.url = url

    async def send_request(
        self,
        payload: Dict[str, Any],
        signature_secret: str,
        additional_payload_values: Dict[str, Any] = {},
    ) -> ExternalActionResponse:
        encoded_payload = json.dumps({"payload": payload} | additional_payload_values).encode(
            "utf-8"
        )

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-vocode-signature": self._encode_payload(encoded_payload, signature_secret),
        }

        transport = httpx.AsyncHTTPTransport(retries=2)
        async with httpx.AsyncClient(
            headers=headers,
            transport=transport,
            timeout=10,
        ) as client:
            try:
                response = await client.post(
                    self.url,
                    content=encoded_payload,
                )
                response.raise_for_status()
                data = response.json()
                return self._validate_response(data)
            except httpx.HTTPStatusError as e:
                logger.error(f"[External Actions] Request failed: {e}")
                if e.response.status_code == 401:
                    return ExternalActionResponse(
                        result={"info": ExternalActionsErrorResponses.unauthorized},
                        success=False,
                    )
                elif e.response.status_code == 403:
                    return ExternalActionResponse(
                        result={"info": ExternalActionsErrorResponses.forbidden},
                        success=False,
                    )
                if 400 <= e.response.status_code < 500:
                    return ExternalActionResponse(
                        result={"info": ExternalActionsErrorResponses.client_error},
                        success=False,
                    )
                elif e.response.status_code >= 500:
                    return ExternalActionResponse(
                        result={
                            "info": ExternalActionsErrorResponses.server_error.format(
                                status=e.response.status_code, text=e.response.text
                            )
                        },
                        success=False,
                    )
                else:
                    raise e
            except ExternalActionValueError as e:
                return ExternalActionResponse(
                    result={"info": ExternalActionsErrorResponses.input_error.format(error=str(e))},
                    success=False,
                )

    def _encode_payload(self, payload: bytes, signature_secret: str) -> str:
        signature_as_bytes = base64.b64decode(signature_secret)
        digest = hmac.new(signature_as_bytes, payload, hashlib.sha256).digest()
        return base64.b64encode(digest).decode()

    def _validate_response(self, response: Dict[str, Any]) -> ExternalActionResponse:
        if "result" not in response:
            raise ExternalActionValueError("Invalid response format: missing 'result'")
        if not isinstance(response["result"], dict):
            raise ExternalActionValueError(
                "Invalid response format: 'agent_message' must be a key value map"
            )
        if "agent_message" in response and not isinstance(response["agent_message"], str):
            raise ExternalActionValueError(
                "Invalid response format: 'agent_message' must be a string"
            )
        return ExternalActionResponse(
            result=response["result"],
            agent_message=response.get("agent_message"),
            success=True,
        )



import asyncio
import base64
import json
import os
import threading
from enum import Enum
from typing import AsyncGenerator, Optional

from fastapi import WebSocket
from loguru import logger

from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.events import PhoneCallConnectedEvent
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import PhoneCallDirection, TwilioConfig
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.models.transcript import Message
from vocode.streaming.output_device.twilio_output_device import TwilioOutputDevice
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.synthesizer.base_synthesizer import SynthesisResult
from vocode.streaming.synthesizer.input_streaming_synthesizer import InputStreamingSynthesizer
from vocode.streaming.telephony.client.twilio_client import TwilioClient
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.telephony.conversation.abstract_phone_conversation import (
    AbstractPhoneConversation,
)
from vocode.streaming.telephony.conversation.mark_message_queue import (
    ChunkFinishedMarkMessage,
    MarkMessage,
    MarkMessageQueue,
    UtteranceFinishedMarkMessage,
)
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.utils import create_utterance_id
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log
from vocode.streaming.utils.events_manager import EventsManager
from vocode.streaming.utils.state_manager import TwilioPhoneConversationStateManager


class TwilioPhoneConversationWebsocketAction(Enum):
    CLOSE_WEBSOCKET = 1


class TwilioPhoneConversation(AbstractPhoneConversation[TwilioOutputDevice]):
    telephony_provider = "twilio"

    def __init__(
        self,
        direction: PhoneCallDirection,
        from_phone: str,
        to_phone: str,
        base_url: str,
        config_manager: BaseConfigManager,
        agent_config: AgentConfig,
        transcriber_config: TranscriberConfig,
        synthesizer_config: SynthesizerConfig,
        twilio_sid: str,
        agent_factory: AbstractAgentFactory,
        transcriber_factory: AbstractTranscriberFactory,
        synthesizer_factory: AbstractSynthesizerFactory,
        twilio_config: Optional[TwilioConfig] = None,
        conversation_id: Optional[str] = None,
        events_manager: Optional[EventsManager] = None,
        record_call: bool = False,
        speed_coefficient: float = 1.0,
        noise_suppression: bool = False,  # is currently a no-op
    ):
        super().__init__(
            direction=direction,
            from_phone=from_phone,
            to_phone=to_phone,
            base_url=base_url,
            config_manager=config_manager,
            output_device=TwilioOutputDevice(),
            agent_config=agent_config,
            transcriber_config=transcriber_config,
            synthesizer_config=synthesizer_config,
            conversation_id=conversation_id,
            events_manager=events_manager,
            transcriber_factory=transcriber_factory,
            agent_factory=agent_factory,
            synthesizer_factory=synthesizer_factory,
            speed_coefficient=speed_coefficient,
        )
        self.mark_message_queue: MarkMessageQueue = MarkMessageQueue()
        self.config_manager = config_manager
        self.twilio_config = twilio_config or TwilioConfig(
            account_sid=os.environ["TWILIO_ACCOUNT_SID"],
            auth_token=os.environ["TWILIO_AUTH_TOKEN"],
        )
        self.telephony_client = TwilioClient(
            base_url=self.base_url, maybe_twilio_config=self.twilio_config
        )
        self.twilio_sid = twilio_sid
        self.record_call = record_call

    def create_state_manager(self) -> TwilioPhoneConversationStateManager:
        return TwilioPhoneConversationStateManager(self)

    async def attach_ws_and_start(self, ws: WebSocket):
        super().attach_ws(ws)

        await self._wait_for_twilio_start(ws)
        await self.start()
        self.events_manager.publish_event(
            PhoneCallConnectedEvent(
                conversation_id=self.id,
                to_phone_number=self.to_phone,
                from_phone_number=self.from_phone,
            )
        )
        while self.active:
            message = await ws.receive_text()
            response = await self._handle_ws_message(message)
            if response == TwilioPhoneConversationWebsocketAction.CLOSE_WEBSOCKET:
                break
        await ws.close(code=1000, reason=None)
        await self.terminate()

    async def _wait_for_twilio_start(self, ws: WebSocket):
        assert isinstance(self.output_device, TwilioOutputDevice)
        while True:
            message = await ws.receive_text()
            if not message:
                continue
            data = json.loads(message)
            if data["event"] == "start":
                logger.debug(f"Media WS: Received event '{data['event']}': {message}")
                self.output_device.stream_sid = data["start"]["streamSid"]
                break

    async def _handle_ws_message(self, message) -> Optional[TwilioPhoneConversationWebsocketAction]:
        if message is None:
            return TwilioPhoneConversationWebsocketAction.CLOSE_WEBSOCKET

        data = json.loads(message)
        if data["event"] == "media":
            media = data["media"]
            chunk = base64.b64decode(media["payload"])
            self.receive_audio(chunk)
        if data["event"] == "mark":
            mark_name = data["mark"]["name"]
            if mark_name.startswith("chunk-"):
                utterance_id, chunk_idx = mark_name.split("-")[1:]
                self.mark_message_queue.put_nowait(
                    utterance_id=utterance_id,
                    mark_message=ChunkFinishedMarkMessage(chunk_idx=int(chunk_idx)),
                )
            elif mark_name.startswith("utterance"):
                utterance_id = mark_name.split("-")[1]
                self.mark_message_queue.put_nowait(
                    utterance_id=utterance_id,
                    mark_message=UtteranceFinishedMarkMessage(),
                )
        elif data["event"] == "stop":
            logger.debug(f"Media WS: Received event 'stop': {message}")
            logger.debug("Stopping...")
            return TwilioPhoneConversationWebsocketAction.CLOSE_WEBSOCKET
        return None

    async def _send_chunks(
        self,
        utterance_id: str,
        chunk_generator: AsyncGenerator[SynthesisResult.ChunkResult, None],
        clear_message_lock: asyncio.Lock,
        stop_event: threading.Event,
    ):
        chunk_idx = 0
        try:
            async for chunk_result in chunk_generator:
                async with clear_message_lock:
                    if stop_event.is_set():
                        break
                    self.output_device.consume_nonblocking(chunk_result.chunk)
                    self.output_device.send_chunk_finished_mark(utterance_id, chunk_idx)
                    chunk_idx += 1
        except asyncio.CancelledError:
            pass
        finally:
            logger.debug("Finished sending all chunks to Twilio")
            self.output_device.send_utterance_finished_mark(utterance_id)

    async def send_speech_to_output(
        self,
        message: str,
        synthesis_result: SynthesisResult,
        stop_event: threading.Event,
        seconds_per_chunk: float,
        transcript_message: Optional[Message] = None,
        started_event: Optional[threading.Event] = None,
    ):
        """In contrast with send_speech_to_output in the base class, this function uses mark messages
        to support interruption - we send all chunks to the output device, and then wait for mark messages[0]
        that indicate that each chunk has been played. This means that we don't need to depends on asyncio.sleep
        to support interruptions.

        Once we receive an interruption signal:
        - we send a clear message to Twilio to stop playing all queued audio
        - based on the number of mark messages we've received back, we know how many chunks were played and can indicate on the transcript

        [0] https://www.twilio.com/docs/voice/twiml/stream#websocket-messages-to-twilio
        """

        if self.transcriber.get_transcriber_config().mute_during_speech:
            logger.debug("Muting transcriber")
            self.transcriber.mute()
        message_sent = message
        cut_off = False
        chunk_idx = 0
        seconds_spoken = 0.0
        logger.debug(f"Start sending speech {message} to output")

        utterance_id = create_utterance_id()
        self.mark_message_queue.create_utterance_queue(utterance_id)

        first_chunk_span = self._maybe_create_first_chunk_span(synthesis_result, message)

        clear_message_lock = asyncio.Lock()

        asyncio_create_task_with_done_error_log(
            self._send_chunks(
                utterance_id,
                synthesis_result.chunk_generator,
                clear_message_lock,
                stop_event,
            ),
        )
        mark_event: MarkMessage
        first = True
        while True:
            mark_event = await self.mark_message_queue.get(utterance_id)
            if isinstance(mark_event, UtteranceFinishedMarkMessage):
                break
            if first and first_chunk_span:
                self._track_first_chunk(first_chunk_span, synthesis_result)
            first = False
            seconds_spoken = mark_event.chunk_idx * seconds_per_chunk
            # Lock here so that we check the stop event and send the clear message atomically
            # w.r.t. the _send_chunks task which also checks the stop event
            # Otherwise, we could send the clear message while _send_chunks is in the middle of sending a chunk
            # and the synthesis wouldn't be cleared
            async with clear_message_lock:
                if stop_event.is_set():
                    self.output_device.send_clear_message()
                    logger.debug(
                        "Interrupted, stopping text to speech after {} chunks".format(chunk_idx)
                    )
                    message_sent = synthesis_result.get_message_up_to(seconds_spoken)
                    cut_off = True
                    break
            if chunk_idx == 0:
                if started_event:
                    started_event.set()
            self.mark_last_action_timestamp()
            chunk_idx += 1
            seconds_spoken += seconds_per_chunk
            if transcript_message:
                transcript_message.text = synthesis_result.get_message_up_to(seconds_spoken)
        self.mark_message_queue.delete_utterance_queue(utterance_id)
        if self.transcriber.get_transcriber_config().mute_during_speech:
            logger.debug("Unmuting transcriber")
            self.transcriber.unmute()
        if transcript_message:
            # For input streaming synthesizers, we have to buffer the message as it is streamed in
            # What is said is federated fully by synthesis_result.get_message_up_to
            if isinstance(self.synthesizer, InputStreamingSynthesizer):
                message_sent = transcript_message.text
            else:
                transcript_message.text = message_sent
            transcript_message.is_final = not cut_off
        if synthesis_result.synthesis_total_span:
            synthesis_result.synthesis_total_span.finish()
        return message_sent, cut_off



import functools
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import sentry_sdk
from loguru import logger
from sentry_sdk.tracing import Span, Transaction, _SpanRecorder

from vocode import get_serialized_ctx_wrappers

if TYPE_CHECKING:
    from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer

_SYNTHESIZER_NAMES = {
    "AzureSynthesizer": "azure",
    "ElevenLabsSynthesizer": "eleven_labs",
    "ElevenLabsWSSynthesizer": "eleven_labs_ws",
    "PlayHtSynthesizer": "play_ht",
    "PlayHtSynthesizerV2": "play_ht_v2",
    "RimeSynthesizer": "rime",
}

_FILTERED_SPANS = {"middleware.starlette.receive", "middleware.starlette.send", "Queue.get"}


class CustomSentrySpans:
    CONNECTED_TO_FIRST_SEND = "connected_to_first_send"
    ENDPOINTING_LATENCY = "endpointing_latency"
    FIRST_SEND_TO_FIRST_RECEIVE = "first_send_to_first_receive"
    LANGUAGE_MODEL_TIME_TO_FIRST_TOKEN = "language_model_time_to_first_token"
    LATENCY_OF_CONVERSATION = "latency_of_conversation"
    LATENCY_OF_TRANSCRIPTION_START = "latency_of_transcription_start"
    LLM_FIRST_SENTENCE_TOTAL = "llm_first_sentence_total"
    START_TO_CONNECTION = "start_to_connection"
    SYNTHESIS_GENERATE_FIRST_CHUNK = "synthesis_generate_first_chunk"
    SYNTHESIS_TIME_TO_FIRST_TOKEN = "synthesis_time_to_first_token"
    TIME_TO_FIRST_TOKEN = "time_to_first_token"

    SYNTHESIZER_SYNTHESIS_TOTAL = ".synthesis_total"
    SYNTHESIZER_TIME_TO_FIRST_TOKEN = ".time_to_first_token"
    SYNTHESIZER_CREATE_SPEECH = ".create_speech"

    @classmethod
    def is_present(cls, value):
        for attr in dir(cls):
            if not attr.startswith("__"):
                attr_value = getattr(cls, attr)
                if isinstance(attr_value, str) and (
                    attr_value == value or value.endswith(attr_value)
                ):
                    return True
        return False


class SentryConfiguredContextManager:
    """
    A context manager that only executes a function if Sentry is configured.

    Attributes:
        func (Callable): The function to be executed.
        args (Tuple): The positional arguments to pass to the function.
        kwargs (Dict): The keyword arguments to pass to the function.
        result (Any): The result of the function execution.
    """

    def __init__(self, func: Callable, *args: Tuple, **kwargs: Dict) -> None:
        """
        Constructs all the necessary attributes for the SentryConfiguredContextManager object.

        Args:
            func (Callable): The function to be executed.
            *args (Tuple): The positional arguments to pass to the function.
            **kwargs (Dict): The keyword arguments to pass to the function.
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.result: Optional[Any] = None

    @property
    def is_configured(self) -> bool:
        """
        Checks if Sentry is configured.

        Returns:
            bool: True if Sentry is configured, False otherwise.
        """
        client = sentry_sdk.Hub.current.client
        if client is not None and client.options is not None and "dsn" in client.options:
            return True
        return False

    def __enter__(self) -> Optional[Any]:
        """
        Executes the function if Sentry is configured.

        Returns:
            Any: The result of the function execution, or None if Sentry is not configured.
        """
        if self.is_configured:
            self.result = self.func(*self.args, **self.kwargs)
            return self.result
        else:
            return None

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Defines the behavior when exiting the context.

        Args:
            exc_type (Any): The type of the exception.
            exc_val (Any): The value of the exception.
            exc_tb (Any): The traceback of the exception.
        """
        pass

    def __call__(self) -> Optional[Any]:
        """
        Executes the function if Sentry is configured, and prints a message if it's not.

        Returns:
            Any: The result of the function execution, or None if Sentry is not configured.
        """
        if self.is_configured:
            return self.func(*self.args, **self.kwargs)
        else:
            logger.debug("Sentry is not configured, skipping function execution.")
            return None

    def execute(self) -> Optional[Any]:
        """Executes the wrapped function immediately and returns its result."""
        return self()


def sentry_configured(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        context_manager = SentryConfiguredContextManager(func, *args, **kwargs)
        if "with" not in func.__code__.co_names:
            return context_manager.execute()
        return context_manager

    return wrapper


def synthesizer_base_name_if_should_report_to_sentry(
    synthesizer: "BaseSynthesizer",
) -> Optional[str]:
    """Returns a synthesizer name if we should report metrics to Sentry for this
    kind of synthesizer; else returns None.
    """
    return f"synthesizer.{_SYNTHESIZER_NAMES.get(synthesizer.__class__.__qualname__)}"


@sentry_configured
def set_tags(span: Span) -> Span:
    serialized_ctx_wrappers = get_serialized_ctx_wrappers()
    for name, value in serialized_ctx_wrappers.items():
        span.set_tag(name, value)
    if not span.description:
        span.description = serialized_ctx_wrappers.get("conversation_id") or span.op
    return span


@sentry_configured
def get_span_by_op(op_value):
    transaction: Transaction = sentry_sdk.Hub.current.scope.transaction
    if transaction is not None:
        # Probably not great accessing an internal variable but transaction spans aren't
        # exposed publicly so it is what it is.
        span_matches = [
            span
            for span in transaction._span_recorder.spans
            if span.op == op_value and span.timestamp is None
        ]
        if span_matches:
            most_recent_span = max(
                span_matches, key=lambda span: span.start_timestamp, default=None
            )
            if most_recent_span is not None:
                return set_tags(most_recent_span)
        else:
            # If no span with the matching op was found
            logger.error(f"No span found with op '{op_value}'.")
            return None
    else:
        logger.debug("No active transaction found.")
        return None


@sentry_configured
def complete_span_by_op(op_value):
    span = get_span_by_op(op_value)
    if span is None:
        logger.error(f"No span found with op '{op_value}'.")
        return None
    span.finish()


@sentry_configured
def sentry_create_span(*args, sentry_callable: Callable, **kwargs) -> Span:
    span = sentry_callable(*args, **kwargs)

    return set_tags(span)


class SpanRecorder(_SpanRecorder):

    def __init__(self, maxlen):
        self.maxlen = 900
        self._auto_spans = []
        self._custom_spans = []
        self._low_prio_spans = []

    def add(self, span: Span):
        if span.op in _FILTERED_SPANS and span.description in _FILTERED_SPANS:
            self._low_prio_spans.append(span)
        else:
            if CustomSentrySpans.is_present(span.op):
                self._custom_spans.append(span)
            else:
                self._auto_spans.append(span)

    @property
    def spans(self):
        return (self._custom_spans + self._auto_spans + self._low_prio_spans)[: self.maxlen]


def init_span_recorder(self, maxlen: int):
    if self._span_recorder is None:
        self._span_recorder = SpanRecorder(maxlen)


Span.init_span_recorder = init_span_recorder  # type: ignore



import ctypes


class WhisperFullParams(ctypes.Structure):
    _fields_ = [
        ("strategy", ctypes.c_int),
        #
        ("n_max_text_ctx", ctypes.c_int),
        ("n_threads", ctypes.c_int),
        ("offset_ms", ctypes.c_int),
        ("duration_ms", ctypes.c_int),
        #
        ("translate", ctypes.c_bool),
        ("no_context", ctypes.c_bool),
        ("single_segment", ctypes.c_bool),
        ("print_special", ctypes.c_bool),
        ("print_progress", ctypes.c_bool),
        ("print_realtime", ctypes.c_bool),
        ("print_timestamps", ctypes.c_bool),
        #
        ("token_timestamps", ctypes.c_bool),
        ("thold_pt", ctypes.c_float),
        ("thold_ptsum", ctypes.c_float),
        ("max_len", ctypes.c_int),
        ("split_on_word", ctypes.c_bool),
        ("max_tokens", ctypes.c_int),
        #
        ("speed_up", ctypes.c_bool),
        ("audio_ctx", ctypes.c_int),
        #
        ("initial_prompt", ctypes.c_char_p),
        ("prompt_tokens", ctypes.c_void_p),
        ("prompt_n_tokens", ctypes.c_int),
        #
        ("language", ctypes.c_char_p),
        #
        ("suppress_blank", ctypes.c_bool),
        ("suppress_non_speech_tokens", ctypes.c_bool),
        #
        ("temperature", ctypes.c_float),
        ("max_initial_ts", ctypes.c_float),
        ("length_penalty", ctypes.c_float),
        #
        ("temperature_inc", ctypes.c_float),
        ("entropy_thold", ctypes.c_float),
        ("logprob_thold", ctypes.c_float),
        ("no_speech_thold", ctypes.c_float),
        #
        ("greedy", ctypes.c_int * 1),
        ("beam_search", ctypes.c_int * 3),
        #
        ("new_segment_callback", ctypes.c_void_p),
        ("new_segment_callback_user_data", ctypes.c_void_p),
        #
        ("progress_callback", ctypes.c_void_p),
        ("progress_callback_user_data", ctypes.c_void_p),
        #
        ("encoder_begin_callback", ctypes.c_void_p),
        ("encoder_begin_callback_user_data", ctypes.c_void_p),
        #
        ("logits_filter_callback", ctypes.c_void_p),
        ("logits_filter_callback_user_data", ctypes.c_void_p),
    ]



from abc import ABC, abstractmethod

from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer
from vocode.streaming.utils.async_requester import AsyncRequestor


class AbstractSynthesizerFactory(ABC):
    # TODO(DOW-48): Make this not require async_requestor
    @abstractmethod
    def create_synthesizer(
        self,
        synthesizer_config: SynthesizerConfig,
    ) -> BaseSynthesizer:
        pass



import asyncio
import audioop
import base64
import io
import json
from typing import Optional

import aiohttp
from loguru import logger

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import (
    RIME_DEFAULT_REDUCE_LATENCY,
    RIME_DEFAULT_SPEED_ALPHA,
    RimeSynthesizerConfig,
)
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult

# TODO: [OSS] Remove call to internal library with Synthesizers refactor

# https://rime.ai/docs/quickstart

WAV_HEADER_LENGTH = 44


class RimeSynthesizer(BaseSynthesizer[RimeSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: RimeSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        self.base_url = synthesizer_config.base_url
        self.model_id = synthesizer_config.model_id
        self.speaker = synthesizer_config.speaker
        self.speed_alpha = synthesizer_config.speed_alpha
        self.sampling_rate = synthesizer_config.sampling_rate
        self.reduce_latency = synthesizer_config.reduce_latency
        self.api_key = f"Bearer {getenv('RIME_API_KEY')}"

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: RimeSynthesizerConfig):
        return ":".join(
            (
                "rime",
                synthesizer_config.speaker,
                str(synthesizer_config.speed_alpha),
                synthesizer_config.audio_encoding,
            )
        )

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        self.total_chars += len(message.text)
        headers = {
            "Authorization": self.api_key,
            "Content-Type": "application/json",
        }

        body = self.get_request_body(message.text)

        async with self.async_requestor.get_session().post(
            self.base_url,
            headers=headers,
            json=body,
            timeout=aiohttp.ClientTimeout(total=15),
        ) as response:
            if not response.ok:
                raise Exception(f"Rime API error: {response.status}, {await response.text()}")
            data = json.loads(await response.text())

            audio_content = data.get("audioContent")
            output_bytes = base64.b64decode(audio_content)[WAV_HEADER_LENGTH:]

            if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
                output_bytes = audioop.lin2ulaw(output_bytes, 2)

            return SynthesisResult(
                self._chunk_generator(output_bytes, chunk_size),
                lambda seconds: self.get_message_cutoff_from_total_response_length(
                    self.synthesizer_config, message, seconds, len(output_bytes)
                ),
            )

    @staticmethod
    async def _chunk_generator(output_bytes, chunk_size):
        for i in range(0, len(output_bytes), chunk_size):
            if i + chunk_size > len(output_bytes):
                yield SynthesisResult.ChunkResult(output_bytes[i:], True)
            else:
                yield SynthesisResult.ChunkResult(output_bytes[i : i + chunk_size], False)

    async def get_chunks(
        self,
        headers: dict,
        body: dict,
        chunk_size: int,
        chunk_queue: asyncio.Queue[Optional[bytes]],
    ):
        try:
            async_client = self.async_requestor.get_client()
            stream = await async_client.send(
                async_client.build_request(
                    "POST",
                    self.base_url,
                    headers=headers,
                    json=body,
                ),
                stream=True,
            )
            if not stream.is_success:
                error = await stream.aread()
                logger.error(f"Rime API failed: {stream.status_code} {error.decode('utf-8')}")
                raise Exception(f"Rime API returned {stream.status_code} status code")
            async for chunk in stream.aiter_bytes(chunk_size):
                chunk_queue.put_nowait(chunk)
        except asyncio.CancelledError:
            pass
        finally:
            chunk_queue.put_nowait(None)  # treated as sentinel

    def get_request_body(self, text):
        speed_alpha = self.speed_alpha if self.speed_alpha else RIME_DEFAULT_SPEED_ALPHA
        reduce_latency = self.reduce_latency if self.reduce_latency else RIME_DEFAULT_REDUCE_LATENCY

        body = {
            "text": text,
            "speaker": self.speaker,
            "samplingRate": self.sampling_rate,
            "speedAlpha": speed_alpha,
            "reduceLatency": reduce_latency,
        }

        if self.model_id:
            body["modelId"] = self.model_id

        return body



import asyncio
import audioop
import json
from typing import Optional

import numpy as np
import websockets
from loguru import logger

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import GladiaTranscriberConfig, Transcription
from vocode.streaming.models.websocket import AudioMessage
from vocode.streaming.transcriber.base_transcriber import BaseAsyncTranscriber

GLADIA_URL = "wss://api.gladia.io/audio/text/audio-transcription"


class GladiaTranscriber(BaseAsyncTranscriber[GladiaTranscriberConfig]):
    def __init__(
        self,
        transcriber_config: GladiaTranscriberConfig,
        api_key: Optional[str] = None,
    ):
        super().__init__(transcriber_config)
        self.api_key = api_key or getenv("GLADIA_API_KEY")
        if not self.api_key:
            raise Exception(
                "Please set GLADIA_API_KEY environment variable or pass it as a parameter"
            )
        self._ended = False
        if self.transcriber_config.endpointing_config:
            raise Exception("Gladia endpointing config not supported yet")

        self.buffer = bytearray()

    async def ready(self):
        return True

    async def _run_loop(self):
        await self.process()

    def send_audio(self, chunk):
        if self.transcriber_config.audio_encoding == AudioEncoding.MULAW:
            sample_width = 1
            if isinstance(chunk, np.ndarray):
                chunk = chunk.astype(np.int16)
                chunk = chunk.tobytes()
            chunk = audioop.ulaw2lin(chunk, sample_width)

        self.buffer.extend(chunk)

        if (
            len(self.buffer) / (2 * self.transcriber_config.sampling_rate)
        ) >= self.transcriber_config.buffer_size_seconds:
            self.input_queue.put_nowait(self.buffer)
            self.buffer = bytearray()

    def terminate(self):
        self._ended = True
        super().terminate()

    async def process(self):
        async with websockets.connect(GLADIA_URL) as ws:
            await ws.send(
                json.dumps(
                    {
                        "x_gladia_key": self.api_key,
                        "sample_rate": self.transcriber_config.sampling_rate,
                        "encoding": "wav",
                    }
                )
            )

            async def sender(ws):
                while not self._ended:
                    try:
                        data = await asyncio.wait_for(self.input_queue.get(), 5)
                    except asyncio.exceptions.TimeoutError:
                        break

                    await ws.send(
                        json.dumps(
                            {
                                "x_gladia_key": self.api_key,
                                "frames": AudioMessage.from_bytes(data).data,
                            }
                        )
                    )
                logger.debug("Terminating Gladia transcriber sender")

            async def receiver(ws):
                while not self._ended:
                    try:
                        result_str = await ws.recv()
                        data = json.loads(result_str)
                        if "error" in data and data["error"]:
                            raise Exception(data["error"])
                    except websockets.exceptions.ConnectionClosedError as e:
                        logger.debug(e)
                        break

                    if data:
                        is_final = data["type"] == "final"

                        if "transcription" in data and data["transcription"]:
                            self.output_queue.put_nowait(
                                Transcription(
                                    message=data["transcription"],
                                    confidence=data["confidence"],
                                    is_final=is_final,
                                )
                            )

            await asyncio.gather(sender(ws), receiver(ws))



import os
from typing import Dict, Optional

import aiohttp
from loguru import logger

from vocode.streaming.models.telephony import TwilioConfig
from vocode.streaming.telephony.client.abstract_telephony_client import AbstractTelephonyClient
from vocode.streaming.telephony.templater import get_connection_twiml
from vocode.streaming.utils.async_requester import AsyncRequestor


class TwilioBadRequestException(ValueError):
    pass


class TwilioClient(AbstractTelephonyClient):
    def __init__(
        self,
        base_url: str,
        maybe_twilio_config: Optional[TwilioConfig] = None,
    ):
        self.twilio_config = maybe_twilio_config or TwilioConfig(
            account_sid=os.environ["TWILIO_ACCOUNT_SID"],
            auth_token=os.environ["TWILIO_AUTH_TOKEN"],
        )
        self.auth = aiohttp.BasicAuth(
            login=self.twilio_config.account_sid,
            password=self.twilio_config.auth_token,
        )
        super().__init__(base_url=base_url)

    def get_telephony_config(self):
        return self.twilio_config

    async def create_call(
        self,
        conversation_id: str,
        to_phone: str,
        from_phone: str,
        record: bool = False,  # currently no-op
        digits: Optional[str] = None,  # currently no-op
        telephony_params: Optional[Dict[str, str]] = None,
    ) -> str:
        data = {
            "Twiml": self.get_connection_twiml(conversation_id=conversation_id).body.decode(
                "utf-8"
            ),
            "To": f"+{to_phone}",
            "From": f"+{from_phone}",
            **(telephony_params or {}),
        }
        if digits:
            data["SendDigits"] = digits
        async with AsyncRequestor().get_session().post(
            f"https://api.twilio.com/2010-04-01/Accounts/{self.twilio_config.account_sid}/Calls.json",
            auth=self.auth,
            data=data,
        ) as response:
            if not response.ok:
                if response.status == 400:
                    logger.error(
                        f"Failed to create call: {response.status} {response.reason} {await response.json()}"
                    )
                    raise TwilioBadRequestException(
                        "Telephony provider rejected call; this is usually due to a bad/malformed number. "
                        "If this persists, and you're sure that the number is well-formed, "
                        "please contact us."
                    )
                raise RuntimeError(f"Failed to create call: {response.status} {response.reason}")
            response = await response.json()
            return response["sid"]

    def get_connection_twiml(self, conversation_id: str):
        return get_connection_twiml(call_id=conversation_id, base_url=self.base_url)

    async def end_call(self, twilio_sid):
        async with AsyncRequestor().get_session().post(
            f"https://api.twilio.com/2010-04-01/Accounts/{self.twilio_config.account_sid}/Calls/{twilio_sid}.json",
            auth=self.auth,
            data={"Status": "completed"},
        ) as response:
            if not response.ok:
                raise RuntimeError(f"Failed to end call: {response.status} {response.reason}")
            response = await response.json()
            return response["status"] == "completed"



from __future__ import annotations

import asyncio
import json
import random
import typing
from enum import Enum
from typing import TYPE_CHECKING, AsyncGenerator, Dict, Generic, Optional, Tuple, TypeVar, Union

import sentry_sdk
from loguru import logger
from pydantic.v1 import BaseModel

from vocode import sentry_span_tags
from vocode.streaming.action.abstract_factory import AbstractActionFactory
from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.action.default_factory import DefaultActionFactory
from vocode.streaming.action.execute_external_action import ExecuteExternalActionVocodeActionConfig
from vocode.streaming.action.phone_call_action import (
    TwilioPhoneConversationAction,
    VonagePhoneConversationAction,
)
from vocode.streaming.agent.goodbye import is_goodbye_simple
from vocode.streaming.agent.phrase_trigger import matches_phrase_trigger
from vocode.streaming.models.actions import (
    ActionConfig,
    ActionInput,
    ActionOutput,
    EndOfTurn,
    FunctionCall,
)
from vocode.streaming.models.agent import AgentConfig, ChatGPTAgentConfig, LLMAgentConfig
from vocode.streaming.models.events import Sender
from vocode.streaming.models.message import BaseMessage, BotBackchannel, SilenceMessage
from vocode.streaming.models.model import TypedModel
from vocode.streaming.models.transcriber import Transcription
from vocode.streaming.models.transcript import Message, Transcript
from vocode.streaming.utils import unrepeating_randomizer
from vocode.streaming.utils.speed_manager import SpeedManager
from vocode.streaming.utils.worker import (
    InterruptibleAgentResponseEvent,
    InterruptibleEvent,
    InterruptibleEventFactory,
    InterruptibleWorker,
)
from vocode.utils.sentry_utils import CustomSentrySpans, sentry_create_span

if TYPE_CHECKING:
    from vocode.streaming.utils.state_manager import AbstractConversationStateManager

AGENT_TRACE_NAME = "agent"
POST_QUESTION_BACKCHANNELS = [
    "Oh okay, got it.",
    "Oh, okay, got it.",
    "Oh okay got it.",
    "Ohhh, okay, got it.",
    "Oh got it, okay.",
    "Ohhh got it, okay.",
    "Ohhh, got it, okay.",
    "Oh, got it, okay.",
]


class AgentInputType(str, Enum):
    BASE = "agent_input_base"
    TRANSCRIPTION = "agent_input_transcription"
    ACTION_RESULT = "agent_input_action_result"


class AgentInput(TypedModel, type=AgentInputType.BASE.value):  # type: ignore
    conversation_id: str
    vonage_uuid: Optional[str]
    twilio_sid: Optional[str]
    agent_response_tracker: Optional[asyncio.Event] = None

    class Config:
        arbitrary_types_allowed = True


class TranscriptionAgentInput(AgentInput, type=AgentInputType.TRANSCRIPTION.value):  # type: ignore
    transcription: Transcription


class ActionResultAgentInput(AgentInput, type=AgentInputType.ACTION_RESULT.value):  # type: ignore
    action_input: ActionInput
    action_output: ActionOutput
    is_quiet: bool = False


class AgentResponseType(str, Enum):
    BASE = "agent_response_base"
    MESSAGE = "agent_response_message"
    STOP = "agent_response_stop"
    FILLER_AUDIO = "agent_response_filler_audio"


class AgentResponse(TypedModel, type=AgentResponseType.BASE.value):  # type: ignore
    pass


class AgentResponseMessage(AgentResponse, type=AgentResponseType.MESSAGE.value):  # type: ignore
    message: Union[BaseMessage, EndOfTurn]
    is_interruptible: bool = True
    # Whether the message is the first message in the response; has metrics implications
    is_first: bool = False
    # If the response is not being chunked up into multiple sentences, this is set to True
    is_sole_text_chunk: bool = False


class AgentResponseStop(AgentResponse, type=AgentResponseType.STOP.value):  # type: ignore
    pass


class AgentResponseFillerAudio(
    AgentResponse,
    type=AgentResponseType.FILLER_AUDIO.value,  # type: ignore
):
    pass


class GeneratedResponse(BaseModel):
    message: Union[BaseMessage, FunctionCall, EndOfTurn]
    is_interruptible: bool
    streamed: bool = False


class StreamedResponse(GeneratedResponse):
    streamed: bool = True


AgentConfigType = TypeVar("AgentConfigType", bound=AgentConfig)


class AbstractAgent(Generic[AgentConfigType]):
    def __init__(self, agent_config: AgentConfigType):
        self.agent_config = agent_config

    def get_agent_config(self) -> AgentConfig:
        return self.agent_config

    def update_last_bot_message_on_cut_off(self, message: str):
        """Updates the last bot message in the conversation history when the human cuts off the bot's response."""
        pass

    def get_cut_off_response(self) -> str:
        assert isinstance(self.agent_config, LLMAgentConfig) or isinstance(
            self.agent_config,
            ChatGPTAgentConfig,
        ), "Set cutoff response is only implemented in LLMAgent and ChatGPTAgent"
        assert self.agent_config.cut_off_response is not None
        on_cut_off_messages = self.agent_config.cut_off_response.messages
        assert len(on_cut_off_messages) > 0
        return random.choice(on_cut_off_messages).text


class BaseAgent(AbstractAgent[AgentConfigType], InterruptibleWorker):
    def __init__(
        self,
        agent_config: AgentConfigType,
        action_factory: AbstractActionFactory = DefaultActionFactory(),
        interruptible_event_factory: InterruptibleEventFactory = InterruptibleEventFactory(),
    ):
        self.input_queue: asyncio.Queue[InterruptibleEvent[AgentInput]] = asyncio.Queue()
        self.output_queue: asyncio.Queue[InterruptibleAgentResponseEvent[AgentResponse]] = (
            asyncio.Queue()
        )
        AbstractAgent.__init__(self, agent_config=agent_config)
        InterruptibleWorker.__init__(
            self,
            input_queue=self.input_queue,
            output_queue=self.output_queue,
            interruptible_event_factory=interruptible_event_factory,
        )
        self.action_factory = action_factory
        self.actions_queue: asyncio.Queue[InterruptibleEvent[ActionInput]] = asyncio.Queue()
        self.transcript: Optional[Transcript] = None

        self.functions = self.get_functions() if self.agent_config.actions else None
        self.is_muted = False

        self.post_question_bot_backchannel_randomizer = unrepeating_randomizer(
            POST_QUESTION_BACKCHANNELS,
        )

    def get_functions(self):
        raise NotImplementedError

    def attach_transcript(self, transcript: Transcript):
        self.transcript = transcript

    def attach_conversation_state_manager(
        self,
        conversation_state_manager: AbstractConversationStateManager,
    ):
        self.conversation_state_manager = conversation_state_manager

    def attach_speed_manager(self, speed_manager: SpeedManager):
        self.speed_manager = speed_manager

    def _get_speed_adjusted_silence_seconds(self, seconds: float) -> float:
        speed_coefficient = (
            self.speed_manager.get_speed_coefficient() if self.speed_manager is not None else 1.0
        )
        return seconds / speed_coefficient

    def set_interruptible_event_factory(self, factory: InterruptibleEventFactory):
        self.interruptible_event_factory = factory

    def get_input_queue(
        self,
    ) -> asyncio.Queue[InterruptibleEvent[AgentInput]]:
        return self.input_queue

    def get_output_queue(
        self,
    ) -> asyncio.Queue[InterruptibleAgentResponseEvent[AgentResponse]]:
        return self.output_queue

    def is_first_response(self):
        assert self.transcript is not None

        num_bot_messages = sum(
            1
            for event_log in self.transcript.event_logs
            if isinstance(event_log, Message) and event_log.sender == Sender.BOT
        )

        return num_bot_messages <= (1 if self.agent_config.initial_message is not None else 0)


class RespondAgent(BaseAgent[AgentConfigType]):
    async def _maybe_prepend_interrupt_responses(
        self,
        transcription: Transcription,
        responses_stream: AsyncGenerator[GeneratedResponse, None],
    ) -> AsyncGenerator[GeneratedResponse, None]:
        if transcription.is_interrupt:
            if self.agent_config.cut_off_response:
                cut_off_response = self.get_cut_off_response()
                yield GeneratedResponse(
                    message=BaseMessage(text=cut_off_response),
                    is_interruptible=False,
                )
                return
            if transcription.bot_was_in_medias_res:
                silence_message = SilenceMessage()
                silence_message.trailing_silence_seconds = self._get_speed_adjusted_silence_seconds(
                    silence_message.trailing_silence_seconds
                )
                yield GeneratedResponse(message=silence_message, is_interruptible=True)
        async for response in responses_stream:
            yield response

    async def handle_generate_response(
        self,
        transcription: Transcription,
        agent_input: AgentInput,
    ) -> bool:
        conversation_id = agent_input.conversation_id
        responses = self._maybe_prepend_interrupt_responses(
            transcription=transcription,
            responses_stream=self.generate_response(
                transcription.message,
                is_interrupt=transcription.is_interrupt,
                conversation_id=conversation_id,
                bot_was_in_medias_res=transcription.bot_was_in_medias_res,
            ),
        )
        is_first_response_of_turn = True
        function_call = None

        responses_buffer = ""
        end_of_turn_agent_response_tracker = None

        async for generated_response in responses:
            if is_first_response_of_turn:
                message_type = "UNKNOWN"
                match generated_response.message:
                    case SilenceMessage():  # type: ignore[misc]
                        message_type = "silence"
                    case BotBackchannel():  # type: ignore[misc]
                        message_type = "backchannel"
                    case BaseMessage():  # type: ignore[misc]
                        message_type = "message"
                    case FunctionCall():  # type: ignore[misc]
                        message_type = "function_call"
                    case _:
                        logger.warning(
                            "Unknown message type received for Sentry metrics "
                            f"reporting: {type(generated_response.message)}",
                        )
                span_tags = sentry_span_tags.value
                if span_tags:
                    span_tags["message_type"] = message_type
                    sentry_span_tags.set(span_tags)

            if isinstance(generated_response.message, FunctionCall):
                function_call = generated_response.message
                continue

            agent_response_tracker = agent_input.agent_response_tracker or asyncio.Event()
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(
                    message=generated_response.message,
                    is_first=is_first_response_of_turn,
                ),
                is_interruptible=self.agent_config.allow_agent_to_be_cut_off
                and generated_response.is_interruptible,
                agent_response_tracker=agent_response_tracker,
            )
            if isinstance(generated_response.message, BaseMessage):
                responses_buffer = f"{responses_buffer} {generated_response.message.text}"
            elif isinstance(generated_response.message, EndOfTurn):
                end_of_turn_agent_response_tracker = agent_response_tracker

            if self.agent_config.end_conversation_on_goodbye and isinstance(
                generated_response.message,
                BaseMessage,
            ):
                if is_goodbye_simple(
                    message=generated_response.message.text,
                    phrases=self.agent_config.goodbye_phrases,
                ):
                    logger.debug("Simple goodbye detected, ending conversation")
                    return True
            is_first_response_of_turn = False

        # if the client (the implemented agent) doesn't create an EndOfTurn, then we need to create one
        if not end_of_turn_agent_response_tracker:
            end_of_turn_agent_response_tracker = (
                agent_input.agent_response_tracker or asyncio.Event()
            )
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(
                    message=EndOfTurn(),
                    is_first=is_first_response_of_turn,
                ),
                is_interruptible=self.agent_config.allow_agent_to_be_cut_off
                and generated_response.is_interruptible,
                agent_response_tracker=end_of_turn_agent_response_tracker,
            )

        phrase_trigger_match = (
            matches_phrase_trigger(responses_buffer, self.agent_config.actions)
            if self.agent_config.actions
            else None
        )
        if phrase_trigger_match:
            action_config = self._get_action_config(phrase_trigger_match)
            assert action_config is not None
            action = self.action_factory.create_action(action_config)
            action_input = self.create_action_input(
                action,
                agent_input,
                {},
                end_of_turn_agent_response_tracker,
            )
            self.enqueue_action_input(action, action_input, agent_input.conversation_id)

        # TODO: implement should_stop for generate_responses
        if function_call and self.agent_config.actions is not None:
            await self.call_function(function_call, agent_input)
        return False

    async def handle_respond(self, transcription: Transcription, conversation_id: str) -> bool:
        try:
            response, should_stop = await self.respond(
                transcription.message,
                is_interrupt=transcription.is_interrupt,
                conversation_id=conversation_id,
            )
        except Exception as e:
            logger.error(f"Error while generating response: {e}", exc_info=True)
            response = None
            return True
        if response:
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(message=BaseMessage(text=response)),
                is_interruptible=self.agent_config.allow_agent_to_be_cut_off,
            )
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(message=EndOfTurn()),
                is_interruptible=self.agent_config.allow_agent_to_be_cut_off,
            )
            return should_stop
        else:
            logger.debug("No response generated")
        return False

    async def process(self, item: InterruptibleEvent[AgentInput]):
        assert self.transcript is not None
        try:
            agent_input = item.payload
            if isinstance(agent_input, TranscriptionAgentInput):
                transcription = typing.cast(TranscriptionAgentInput, agent_input).transcription
                self.transcript.add_human_message(
                    text=transcription.message,
                    conversation_id=agent_input.conversation_id,
                )
            elif isinstance(agent_input, ActionResultAgentInput):
                self.transcript.add_action_finish_log(
                    action_input=agent_input.action_input,
                    action_output=agent_input.action_output,
                    conversation_id=agent_input.conversation_id,
                )
                if agent_input.is_quiet:
                    # Do not generate a response to quiet actions
                    logger.debug("Action is quiet, skipping response generation")
                    return
                if agent_input.action_output.canned_response is not None:
                    self.produce_interruptible_agent_response_event_nonblocking(
                        AgentResponseMessage(
                            message=agent_input.action_output.canned_response,
                            is_sole_text_chunk=True,
                        ),
                        is_interruptible=True,
                    )
                    self.produce_interruptible_agent_response_event_nonblocking(
                        AgentResponseMessage(message=EndOfTurn()),
                    )
                    return
                transcription = Transcription(
                    message=agent_input.action_output.response.json(),
                    confidence=1.0,
                    is_final=True,
                )
            else:
                raise ValueError("Invalid AgentInput type")

            if self.is_muted:
                logger.debug("Agent is muted, skipping processing")
                return

            if self.agent_config.send_filler_audio:
                self.produce_interruptible_agent_response_event_nonblocking(
                    AgentResponseFillerAudio(),
                )

            logger.debug("Responding to transcription")
            should_stop = False
            if self.agent_config.generate_responses:
                # TODO (EA): this is quite ugly but necessary to have the agent act properly after an action completes
                if not isinstance(agent_input, ActionResultAgentInput):
                    sentry_create_span(
                        sentry_callable=sentry_sdk.start_span,
                        op=CustomSentrySpans.LANGUAGE_MODEL_TIME_TO_FIRST_TOKEN,
                    )
                should_stop = await self.handle_generate_response(transcription, agent_input)
            else:
                should_stop = await self.handle_respond(transcription, agent_input.conversation_id)

            if should_stop:
                logger.debug("Agent requested to stop")
                self.produce_interruptible_agent_response_event_nonblocking(AgentResponseStop())
                return
        except asyncio.CancelledError:
            pass

    def _get_action_config(self, function_name: str) -> Optional[ActionConfig]:
        if self.agent_config.actions is None:
            return None
        for action_config in self.agent_config.actions:
            if action_config.type == function_name or (
                isinstance(action_config, ExecuteExternalActionVocodeActionConfig)
                and action_config.name == function_name
            ):
                return action_config
        return None

    async def call_function(self, function_call: FunctionCall, agent_input: AgentInput):
        action_config = self._get_action_config(function_call.name)
        if action_config is None:
            logger.error(f"Function {function_call.name} not found in agent config, skipping")
            return
        action = self.action_factory.create_action(action_config)
        params = json.loads(function_call.arguments)
        user_message_tracker = None
        if "user_message" in params:
            user_message = params["user_message"]
            user_message_tracker = asyncio.Event()
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(
                    message=BaseMessage(text=user_message),
                    is_sole_text_chunk=True,
                ),
                is_interruptible=action.is_interruptible,
            )
            self.produce_interruptible_agent_response_event_nonblocking(
                AgentResponseMessage(message=EndOfTurn()),
                agent_response_tracker=user_message_tracker,
            )
        action_input = self.create_action_input(action, agent_input, params, user_message_tracker)
        self.enqueue_action_input(action, action_input, agent_input.conversation_id)

    def create_action_input(
        self,
        action: BaseAction,
        agent_input: AgentInput,
        params: Dict,
        user_message_tracker: Optional[asyncio.Event] = None,
    ) -> ActionInput:
        action_input: ActionInput
        if isinstance(action, VonagePhoneConversationAction):
            assert (
                agent_input.vonage_uuid is not None
            ), "Cannot use VonagePhoneConversationActionFactory unless the attached conversation is a VonagePhoneConversation"
            action_input = action.create_phone_conversation_action_input(
                agent_input.conversation_id,
                params,
                agent_input.vonage_uuid,
                user_message_tracker,
            )
        elif isinstance(action, TwilioPhoneConversationAction):
            assert (
                agent_input.twilio_sid is not None
            ), "Cannot use TwilioPhoneConversationActionFactory unless the attached conversation is a TwilioPhoneConversation"
            action_input = action.create_phone_conversation_action_input(
                agent_input.conversation_id,
                params,
                agent_input.twilio_sid,
                user_message_tracker,
            )
        else:
            action_input = action.create_action_input(
                agent_input.conversation_id,
                params,
                user_message_tracker,
            )
        return action_input

    def enqueue_action_input(
        self,
        action: BaseAction,
        action_input: ActionInput,
        conversation_id: str,
    ):
        event = self.interruptible_event_factory.create_interruptible_event(
            action_input,
            is_interruptible=action.is_interruptible,
        )
        assert self.transcript is not None
        self.transcript.add_action_start_log(
            action_input=action_input,
            conversation_id=conversation_id,
        )
        self.actions_queue.put_nowait(event)

    async def respond(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
    ) -> Tuple[Optional[str], bool]:
        raise NotImplementedError

    def generate_response(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
        bot_was_in_medias_res: bool = False,
    ) -> AsyncGenerator[
        GeneratedResponse,
        None,
    ]:  # tuple of the content and whether it is interruptible
        raise NotImplementedError



# type: ignore
# to be deprecated

from typing import Union

from ..models.agent import RESTfulAgentEnd, RESTfulAgentInput, RESTfulAgentOutput, RESTfulAgentText
from .base_agent import BaseAgent


class RESTfulAgent(BaseAgent):
    def __init__(self):
        super().__init__()
        self.app.post("/respond")(self.respond_rest)

    async def respond(self, human_input, conversation_id) -> RESTfulAgentOutput:
        raise NotImplementedError

    async def respond_rest(
        self, request: RESTfulAgentInput
    ) -> Union[RESTfulAgentText, RESTfulAgentEnd]:
        response = await self.respond(request.human_input, request.conversation_id)
        return response



from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.agent.anthropic_agent import AnthropicAgent
from vocode.streaming.agent.base_agent import BaseAgent
from vocode.streaming.agent.chat_gpt_agent import ChatGPTAgent
from vocode.streaming.agent.echo_agent import EchoAgent
from vocode.streaming.agent.restful_user_implemented_agent import RESTfulUserImplementedAgent
from vocode.streaming.models.agent import (
    AgentConfig,
    AnthropicAgentConfig,
    ChatGPTAgentConfig,
    EchoAgentConfig,
    RESTfulUserImplementedAgentConfig,
)


class DefaultAgentFactory(AbstractAgentFactory):
    def create_agent(self, agent_config: AgentConfig) -> BaseAgent:
        if isinstance(agent_config, ChatGPTAgentConfig):
            return ChatGPTAgent(agent_config=agent_config)
        elif isinstance(agent_config, EchoAgentConfig):
            return EchoAgent(agent_config=agent_config)
        elif isinstance(agent_config, RESTfulUserImplementedAgentConfig):
            return RESTfulUserImplementedAgent(agent_config=agent_config)
        elif isinstance(agent_config, AnthropicAgentConfig):
            return AnthropicAgent(agent_config=agent_config)
        raise Exception("Invalid agent config", agent_config.type)



from vocode.streaming.models.audio import AudioEncoding


class BaseOutputDevice:
    def __init__(self, sampling_rate: int, audio_encoding: AudioEncoding):
        self.sampling_rate = sampling_rate
        self.audio_encoding = audio_encoding

    def start(self):
        pass

    def consume_nonblocking(self, chunk: bytes):
        raise NotImplemented

    def terminate(self):
        pass



import asyncio
import wave
from asyncio import Queue

import numpy as np

from .base_output_device import BaseOutputDevice
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.utils.worker import ThreadAsyncWorker


class FileWriterWorker(ThreadAsyncWorker):
    def __init__(self, input_queue: Queue, wave) -> None:
        super().__init__(input_queue)
        self.wav = wave

    def _run_loop(self):
        while True:
            try:
                block = self.input_janus_queue.sync_q.get()
                self.wav.writeframes(block)
            except asyncio.CancelledError:
                return

    def terminate(self):
        super().terminate()
        self.wav.close()


class FileOutputDevice(BaseOutputDevice):
    DEFAULT_SAMPLING_RATE = 44100

    def __init__(
        self,
        file_path: str,
        sampling_rate: int = DEFAULT_SAMPLING_RATE,
        audio_encoding: AudioEncoding = AudioEncoding.LINEAR16,
    ):
        super().__init__(sampling_rate, audio_encoding)
        self.blocksize = self.sampling_rate
        self.queue: Queue[np.ndarray] = Queue()

        wav = wave.open(file_path, "wb")
        wav.setnchannels(1)  # Mono channel
        wav.setsampwidth(2)  # 16-bit samples
        wav.setframerate(self.sampling_rate)
        self.wav = wav

        self.thread_worker = FileWriterWorker(self.queue, wav)
        self.thread_worker.start()

    def consume_nonblocking(self, chunk):
        chunk_arr = np.frombuffer(chunk, dtype=np.int16)
        for i in range(0, chunk_arr.shape[0], self.blocksize):
            block = np.zeros(self.blocksize, dtype=np.int16)
            size = min(self.blocksize, chunk_arr.shape[0] - i)
            block[:size] = chunk_arr[i : i + size]
            self.queue.put_nowait(block)

    def terminate(self):
        self.thread_worker.terminate()



import asyncio
from typing import Dict, Union

from pydantic.v1 import BaseModel


class ChunkFinishedMarkMessage(BaseModel):
    chunk_idx: int


class UtteranceFinishedMarkMessage(BaseModel):
    pass


MarkMessage = Union[ChunkFinishedMarkMessage, UtteranceFinishedMarkMessage]


class MarkMessageQueue:
    """A keyed asyncio.Queue for MarkMessage objects"""

    def __init__(self):
        self.utterance_queues: Dict[str, asyncio.Queue[MarkMessage]] = {}

    def create_utterance_queue(self, utterance_id: str):
        if utterance_id in self.utterance_queues:
            raise ValueError(f"utterance_id {utterance_id} already exists")
        self.utterance_queues[utterance_id] = asyncio.Queue()

    def put_nowait(
        self,
        utterance_id: str,
        mark_message: MarkMessage,
    ):
        if utterance_id in self.utterance_queues:
            self.utterance_queues[utterance_id].put_nowait(mark_message)

    async def get(
        self,
        utterance_id: str,
    ) -> MarkMessage:
        if utterance_id not in self.utterance_queues:
            raise ValueError(f"utterance_id {utterance_id} not found")
        return await self.utterance_queues[utterance_id].get()

    def delete_utterance_queue(self, utterance_id: str):
        del self.utterance_queues[utterance_id]



import os

from fastapi import Response
from jinja2 import Environment, FileSystemLoader

DEFAULT_TEMPLATE_ENVIRONMENT = Environment(
    loader=FileSystemLoader("%s/templates/" % os.path.dirname(__file__))
)


def render_template(template_name: str, template_environment: Environment, **kwargs):
    template = template_environment.get_template(template_name)
    return template.render(**kwargs)


def get_connection_twiml(
    call_id: str,
    base_url: str,
    template_environment: Environment = DEFAULT_TEMPLATE_ENVIRONMENT,
):
    return Response(
        render_template(
            template_name="twilio_connect_call.xml",
            template_environment=template_environment,
            base_url=base_url,
            id=call_id,
        ),
        media_type="application/xml",
    )



import asyncio
import functools

from loguru import logger

tasks_registry = []


def log_if_exception(reraise_cancelled: bool, future: asyncio.Future) -> None:
    try:
        if exc := future.exception():
            logger.exception(
                f"Vocode wrapped logger; exception raised by task {future}: {exc}",
                exc_info=exc,
            )
    except asyncio.CancelledError:
        if reraise_cancelled:
            raise


def asyncio_create_task_with_done_error_log(
    *args,
    reraise_cancelled: bool = False,
    **kwargs,
) -> asyncio.Task:
    task = asyncio.create_task(*args, **kwargs)
    tasks_registry.append(task)
    task.add_done_callback(functools.partial(log_if_exception, reraise_cancelled))
    task.add_done_callback(lambda t: tasks_registry.remove(t))
    return task



raise DeprecationWarning("This Agent is deprecated and will be removed in the future.")

# OLD CONFIG BELOW
# class LlamacppAgentConfig(AgentConfig, type=AgentType.LLAMACPP.value):  # type: ignore
#     prompt_preamble: str
#     llamacpp_kwargs: dict = {}
#     prompt_template: Optional[Union["PromptTemplate", str]] = None

# ALPACA_TEMPLATE_WITH_HISTORY = """### Instruction:
# Your previous conversation history:
# {history}

# Current instruction/message to respond to: {input}
# ### Response:"""


# class CallbackOutput(BaseModel):
#     finish: bool = False
#     response: Optional[LLMResult] = None
#     token: str = ""


# class FormatHistoryPromptTemplate(PromptTemplate):
#     def format(self, **kwargs: Any) -> str:
#         kwargs = self._merge_partial_and_user_variables(**kwargs)
#         kwargs["history"] = get_buffer_string(kwargs["history"])
#         return DEFAULT_FORMATTER_MAPPING[self.template_format](self.template, **kwargs)


# class CustomStreamingCallbackHandler(BaseCallbackHandler):
#     def __init__(self, output_queue: asyncio.Queue) -> None:
#         super().__init__()
#         self.output_queue = output_queue

#     def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
#         """Run on new LLM token. Only available when streaming is enabled."""
#         self.output_queue.put_nowait(CallbackOutput(token=token))

#     def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
#         """Run when LLM ends running."""
#         self.output_queue.put_nowait(CallbackOutput(finish=True, response=response))


# class LlamacppAgent(RespondAgent[LlamacppAgentConfig]):
#     def __init__(
#         self,
#         agent_config: LlamacppAgentConfig,
#     ):
#         super().__init__(agent_config=agent_config)

#         self.prompt: Union[PromptTemplate, ChatPromptTemplate]
#         if type(agent_config.prompt_template) is str:
#             if agent_config.prompt_template == "alpaca":
#                 self.prompt = FormatHistoryPromptTemplate(
#                     input_variables=["history", "input"],
#                     template=ALPACA_TEMPLATE_WITH_HISTORY,
#                 )
#             else:
#                 raise ValueError(f"Unknown prompt template {agent_config.prompt_template}")
#         else:
#             if agent_config.prompt_template is None:
#                 self.prompt = ChatPromptTemplate.from_messages(
#                     [
#                         MessagesPlaceholder(variable_name="history"),
#                         HumanMessagePromptTemplate.from_template("{input}"),
#                     ]
#                 )
#             else:
#                 self.promt = typing.cast(PromptTemplate, agent_config.prompt_template)

#         self.callback_queue: asyncio.Queue = asyncio.Queue()
#         callback = CustomStreamingCallbackHandler(self.callback_queue)
#         callback_manager = CallbackManager([callback])
#         self.llm = LlamaCpp(callback_manager=callback_manager, **agent_config.llamacpp_kwargs)

#         self.memory = ConversationBufferMemory(return_messages=True)
#         self.memory.chat_memory.messages.append(
#             SystemMessage(content=self.agent_config.prompt_preamble)
#         )

#         self.conversation = ConversationChain(memory=self.memory, prompt=self.prompt, llm=self.llm)
#         self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

#     async def respond(
#         self,
#         human_input,
#         conversation_id: str,
#         is_interrupt: bool = False,
#     ) -> Tuple[str, bool]:
#         text = await asyncio.get_event_loop().run_in_executor(
#             self.thread_pool_executor,
#             lambda input: self.conversation.predict(input=input),
#             human_input,
#         )

#         logger.debug(f"LLM response: {text}")
#         return text, False

#     async def llamacpp_get_tokens(self):
#         while True:
#             callback_output = await self.callback_queue.get()
#             if callback_output.finish:
#                 break
#             yield callback_output.token

#     async def generate_response(
#         self,
#         human_input: str,
#         conversation_id: str,
#         is_interrupt: bool = False,
#     ) -> AsyncGenerator[Tuple[str, bool], None]:
#         asyncio.get_event_loop().run_in_executor(
#             self.thread_pool_executor,
#             lambda input: self.conversation.predict(input=input),
#             human_input,
#         )

#         async for message in collate_response_async(
#             self.llamacpp_get_tokens(),
#         ):
#             yield str(message), True



import os
import random
from typing import Any, AsyncGenerator, Dict, List, Optional, TypeVar, Union

import sentry_sdk
from loguru import logger
from openai import DEFAULT_MAX_RETRIES as OPENAI_DEFAULT_MAX_RETRIES
from openai import AsyncAzureOpenAI, AsyncOpenAI, NotFoundError, RateLimitError

from vocode import sentry_span_tags
from vocode.streaming.action.abstract_factory import AbstractActionFactory
from vocode.streaming.action.default_factory import DefaultActionFactory
from vocode.streaming.agent.base_agent import GeneratedResponse, RespondAgent, StreamedResponse
from vocode.streaming.agent.openai_utils import (
    format_openai_chat_messages_from_transcript,
    openai_get_tokens,
    vector_db_result_to_openai_chat_message,
)
from vocode.streaming.agent.streaming_utils import collate_response_async, stream_response_async
from vocode.streaming.models.actions import FunctionCallActionTrigger
from vocode.streaming.models.agent import ChatGPTAgentConfig
from vocode.streaming.models.events import Sender
from vocode.streaming.models.message import BaseMessage, BotBackchannel, LLMToken
from vocode.streaming.models.transcript import Message
from vocode.streaming.vector_db.factory import VectorDBFactory
from vocode.utils.sentry_utils import CustomSentrySpans, sentry_create_span

ChatGPTAgentConfigType = TypeVar("ChatGPTAgentConfigType", bound=ChatGPTAgentConfig)


def instantiate_openai_client(agent_config: ChatGPTAgentConfig, model_fallback: bool = False):
    if agent_config.azure_params:
        return AsyncAzureOpenAI(
            azure_endpoint=agent_config.azure_params.base_url,
            api_key=agent_config.azure_params.api_key,
            api_version=agent_config.azure_params.api_version,
            max_retries=0 if model_fallback else OPENAI_DEFAULT_MAX_RETRIES,
        )
    else:
        if agent_config.openai_api_key is not None:
            logger.info("Using OpenAI API key override")
        return AsyncOpenAI(
            api_key=agent_config.openai_api_key or os.environ["OPENAI_API_KEY"],
            base_url="https://api.openai.com/v1",
            max_retries=0 if model_fallback else OPENAI_DEFAULT_MAX_RETRIES,
        )


class ChatGPTAgent(RespondAgent[ChatGPTAgentConfigType]):
    openai_client: Union[AsyncOpenAI, AsyncAzureOpenAI]

    def __init__(
        self,
        agent_config: ChatGPTAgentConfigType,
        action_factory: AbstractActionFactory = DefaultActionFactory(),
        vector_db_factory=VectorDBFactory(),
        **kwargs,
    ):
        super().__init__(
            agent_config=agent_config,
            action_factory=action_factory,
            **kwargs,
        )
        self.openai_client = instantiate_openai_client(
            agent_config, model_fallback=agent_config.llm_fallback is not None
        )

        if not self.openai_client.api_key:
            raise ValueError("OPENAI_API_KEY must be set in environment or passed in")

        if self.agent_config.vector_db_config:
            self.vector_db = vector_db_factory.create_vector_db(self.agent_config.vector_db_config)

    def get_functions(self):
        assert self.agent_config.actions
        if not self.action_factory:
            return None
        return [
            self.action_factory.create_action(action_config).get_openai_function()
            for action_config in self.agent_config.actions
            if isinstance(action_config.action_trigger, FunctionCallActionTrigger)
        ]

    def get_chat_parameters(self, messages: Optional[List] = None, use_functions: bool = True):
        assert self.transcript is not None
        is_azure = self._is_azure_model()

        messages = messages or format_openai_chat_messages_from_transcript(
            self.transcript,
            self.get_model_name_for_tokenizer(),
            self.functions,
            self.agent_config.prompt_preamble,
        )

        parameters: Dict[str, Any] = {
            "messages": messages,
            "max_tokens": self.agent_config.max_tokens,
            "temperature": self.agent_config.temperature,
        }

        if is_azure:
            assert self.agent_config.azure_params is not None
            parameters["model"] = self.agent_config.azure_params.deployment_name
        else:
            parameters["model"] = self.agent_config.model_name

        if use_functions and self.functions:
            parameters["functions"] = self.functions

        return parameters

    def _is_azure_model(self) -> bool:
        return self.agent_config.azure_params is not None

    def get_model_name_for_tokenizer(self):
        if not self.agent_config.azure_params:
            return self.agent_config.model_name
        else:
            return self.agent_config.azure_params.openai_model_name

    def apply_model_fallback(self, chat_parameters: Dict[str, Any]):
        if self.agent_config.llm_fallback is None:
            return
        if self.agent_config.llm_fallback.provider == "openai":
            self.agent_config.model_name = self.agent_config.llm_fallback.model_name
            if isinstance(self.openai_client, AsyncAzureOpenAI):
                self.agent_config.azure_params = None
        else:
            if self.agent_config.azure_params:
                self.agent_config.azure_params.deployment_name = (
                    self.agent_config.llm_fallback.model_name
                )
                if isinstance(self.openai_client, AsyncOpenAI):
                    # TODO: handle OpenAI fallback to Azure
                    pass

        self.openai_client = instantiate_openai_client(self.agent_config, model_fallback=False)
        chat_parameters["model"] = self.agent_config.llm_fallback.model_name

    async def _create_openai_stream_with_fallback(
        self, chat_parameters: Dict[str, Any]
    ) -> AsyncGenerator:
        try:
            stream = await self.openai_client.chat.completions.create(**chat_parameters)
        except (NotFoundError, RateLimitError) as e:
            logger.error(
                f"{'Model not found' if isinstance(e, NotFoundError) else 'Rate limit error'} for model_name: {chat_parameters.get('model')}. Applying fallback.",
                exc_info=True,
            )
            self.apply_model_fallback(chat_parameters)
            stream = await self.openai_client.chat.completions.create(**chat_parameters)
        except Exception as e:
            logger.error(
                f"Error while hitting OpenAI with chat_parameters: {chat_parameters}",
                exc_info=True,
            )
            raise e
        return stream

    async def _create_openai_stream(self, chat_parameters: Dict[str, Any]) -> AsyncGenerator:
        if self.agent_config.llm_fallback is not None and self.openai_client.max_retries == 0:
            stream = await self._create_openai_stream_with_fallback(chat_parameters)
        else:
            try:
                stream = await self.openai_client.chat.completions.create(**chat_parameters)
            except Exception as e:
                logger.error(
                    f"Error while hitting OpenAI with chat_parameters: {chat_parameters}",
                    exc_info=True,
                )
                raise e
        return stream

    def should_backchannel(self, human_input: str) -> bool:
        return (
            not self.is_first_response()
            and not human_input.strip().endswith("?")
            and random.random() < self.agent_config.backchannel_probability
        )

    def choose_backchannel(self) -> Optional[BotBackchannel]:
        backchannel = None
        if self.transcript is not None:
            last_bot_message: Optional[Message] = None
            for event_log in self.transcript.event_logs[::-1]:
                if isinstance(event_log, Message) and event_log.sender == Sender.BOT:
                    last_bot_message = event_log
                    break
            if last_bot_message and last_bot_message.text.strip().endswith("?"):
                return BotBackchannel(text=self.post_question_bot_backchannel_randomizer())
        return backchannel

    async def generate_response(
        self,
        human_input: str,
        conversation_id: str,
        is_interrupt: bool = False,
        bot_was_in_medias_res: bool = False,
    ) -> AsyncGenerator[GeneratedResponse, None]:
        assert self.transcript is not None

        chat_parameters = {}
        if self.agent_config.vector_db_config:
            try:
                docs_with_scores = await self.vector_db.similarity_search_with_score(
                    self.transcript.get_last_user_message()[1]
                )
                docs_with_scores_str = "\n\n".join(
                    [
                        "Document: "
                        + doc[0].metadata["source"]
                        + f" (Confidence: {doc[1]})\n"
                        + doc[0].lc_kwargs["page_content"].replace(r"\n", "\n")
                        for doc in docs_with_scores
                    ]
                )
                vector_db_result = (
                    f"Found {len(docs_with_scores)} similar documents:\n{docs_with_scores_str}"
                )
                messages = format_openai_chat_messages_from_transcript(
                    self.transcript,
                    self.agent_config.model_name,
                    self.functions,
                    self.agent_config.prompt_preamble,
                )
                messages.insert(-1, vector_db_result_to_openai_chat_message(vector_db_result))
                chat_parameters = self.get_chat_parameters(messages)
            except Exception as e:
                logger.error(f"Error while hitting vector db: {e}", exc_info=True)
                chat_parameters = self.get_chat_parameters()
        else:
            chat_parameters = self.get_chat_parameters()
        chat_parameters["stream"] = True

        openai_chat_messages: List = chat_parameters.get("messages", [])

        backchannelled = "false"
        backchannel: Optional[BotBackchannel] = None
        if (
            self.agent_config.use_backchannels
            and not bot_was_in_medias_res
            and self.should_backchannel(human_input)
        ):
            backchannel = self.choose_backchannel()
        elif self.agent_config.first_response_filler_message and self.is_first_response():
            backchannel = BotBackchannel(text=self.agent_config.first_response_filler_message)

        if backchannel is not None:
            # The LLM needs the backchannel context manually - otherwise we're in a race condition
            # between sending the response and generating ChatGPT's response
            openai_chat_messages.append({"role": "assistant", "content": backchannel.text})
            yield GeneratedResponse(
                message=backchannel,
                is_interruptible=True,
            )
            backchannelled = "true"

        span_tags = sentry_span_tags.value
        if span_tags:
            span_tags["prior_backchannel"] = backchannelled
            sentry_span_tags.set(span_tags)

        first_sentence_total_span = sentry_create_span(
            sentry_callable=sentry_sdk.start_span, op=CustomSentrySpans.LLM_FIRST_SENTENCE_TOTAL
        )

        ttft_span = sentry_create_span(
            sentry_callable=sentry_sdk.start_span, op=CustomSentrySpans.TIME_TO_FIRST_TOKEN
        )

        stream = await self._create_openai_stream(chat_parameters)

        response_generator = collate_response_async
        using_input_streaming_synthesizer = (
            self.conversation_state_manager.using_input_streaming_synthesizer()
        )
        if using_input_streaming_synthesizer:
            response_generator = stream_response_async
        async for message in response_generator(
            conversation_id=conversation_id,
            gen=openai_get_tokens(
                stream,
            ),
            get_functions=True,
            sentry_span=ttft_span,
        ):
            if first_sentence_total_span:
                first_sentence_total_span.finish()

            ResponseClass = (
                StreamedResponse if using_input_streaming_synthesizer else GeneratedResponse
            )
            MessageType = LLMToken if using_input_streaming_synthesizer else BaseMessage
            if isinstance(message, str):
                yield ResponseClass(
                    message=MessageType(text=message),
                    is_interruptible=True,
                )
            else:
                yield ResponseClass(
                    message=message,
                    is_interruptible=True,
                )



from abc import ABC, abstractmethod

from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.models.actions import ActionConfig


class AbstractActionFactory(ABC):
    @abstractmethod
    def create_action(self, action_config: ActionConfig) -> BaseAction:
        pass
